#include <Rinternals.h>
#include <sys/types.h>
#include <limits.h>
#include <db.h>
#include "bitstring.h"

SEXP BNSL_db_tag;

#define as_db(x) ((DB*)R_ExternalPtrAddr((x)))

void add_class(SEXP obj,const char *name) {
    SEXP class, oclass;
    int i, leno;
    
    oclass = getAttrib(obj, R_ClassSymbol);
    leno = length(oclass);

    if( oclass == R_NilValue )
        PROTECT(class = mkString(name));
    else {
        PROTECT(class = allocVector(STRSXP, leno+1));
        for (i = 0; i < leno; i++)
            SET_VECTOR_ELT(class, i, VECTOR_ELT(oclass, i));
        SET_VECTOR_ELT(class, leno, mkChar(name));
    }
    classgets(obj, class);
    UNPROTECT(1);
}
void BNSL_FinalizeDB(SEXP d) {
  DB *db = (DB*)R_ExternalPtrAddr(d);
  if(db == NULL) return;
  db->close(db,0);
}

void get_db_info(SEXP _d,DB **db,char **K,int *n) {
  SEXP  d;
  SEXP _k;

  d = VECTOR_ELT(_d,0);
  if(TYPEOF(d) != EXTPTRSXP || R_ExternalPtrTag(d) != BNSL_db_tag)
    error("object passed is not a valid database");
  _k = R_ExternalPtrProtected(VECTOR_ELT(_d,0));
  (*db)   = as_db(VECTOR_ELT(_d,0));
  (*K) = CHAR(_k)+1;
  (*n) = CHAR(_k)[0];
}

SEXP BNSL_MakeDB(SEXP N,SEXP file,SEXP cache) {
  DB* db;
  SEXP r,p,key;
  db_create(&db,NULL,0);
  if(file != R_NilValue)
    db->set_cachesize(db,INTEGER(cache)[0],INTEGER(cache)[1],INTEGER(cache)[2]);
  db->open(db,NULL,
	   file == R_NilValue ? NULL : CHAR(STRING_ELT(file,0)),
	   NULL,DB_HASH,DB_CREATE,0);

  PROTECT(r   = allocVector(VECSXP,1));
  PROTECT(key = allocVector(RAWSXP,bitstr_size(INTEGER(N)[0])+2));
  RAW(key)[0] = (unsigned char)(INTEGER(N)[0]); /* Pascal-style string! */

  PROTECT(p = R_MakeExternalPtr(db,BNSL_db_tag,key));
  R_RegisterCFinalizer(p,BNSL_FinalizeDB);
  SET_VECTOR_ELT(r,0,p);
  setAttrib(r,install("vars"),duplicate(N));
  setAttrib(r,install("file"),duplicate(file));
  add_class(r,"db");

  UNPROTECT(3);
  return r;
}
void print_key(SEXP db) {
  int  i;
  SEXP key = R_ExternalPtrProtected(db);
  char *K  = CHAR(key);
  int   s  = K[0];K++;
  for(i=0;i<s;i++) 
    printf("%d",bit_test(K,i) ? 1 : 0);
  printf("\n");
}

void update_key(SEXP db,SEXP child,SEXP set,DBT *k) {
  int  i,s;
  char *K;
  SEXP key = R_ExternalPtrProtected(db);

  K = CHAR(key);
  s = K[0];K++;k->data = K;
  K[0] = INTEGER(child)[0];K++;
  memset(K,0,sizeof(char)*bitstr_size(s));
  bit_nclear(K,0,s);
  for(i=0;i<length(set);i++)
    if(INTEGER(set)[i] != INTEGER(child)[0]) bit_set(K,INTEGER(set)[i]-1);
  k->size = sizeof(char)*(bitstr_size(s)+1);
  k->flags= 0;
}

void set_key(SEXP db,int child,SEXP set,DBT*k) {
  int i,s;
  
}



void report_key_error(int e) {
  switch(e) {
  case EACCES:
    error("attempted to write a read-only database");
  case EINVAL:
    error("invalid key/data combination");
  case DB_DONOTINDEX:
    error("\"Null\" return from secondary callback");
  case DB_KEYEMPTY:
    error("Key/data deleted for never existed");
  case DB_KEYEXIST:
    error("The key/data pair already exists");
  case DB_LOCK_DEADLOCK:
    error("deadlock");
  case DB_LOCK_NOTGRANTED:
    error("lock unavailable");
  case DB_NOSERVER:
    error("Server panic return");
  case DB_NOSERVER_HOME:
    error("Bad home sent to server");
  case DB_NOSERVER_ID:
    error("Bad ID sent to server");
  case DB_OLD_VERSION:
    error("Out-of-date version");
  case DB_PAGE_NOTFOUND:
    error("Requested page not found");
  default:
    error("unknown error");
  }
}

SEXP BNSL_Sync(SEXP db) {
  DB* m = as_db(VECTOR_ELT(db,0));
  m->sync(m,0);
  return R_NilValue;
}


SEXP BNSL_PutDB(SEXP db,SEXP child,SEXP set,SEXP ovwr,SEXP value) {
  int  i,s;
  DBT  k,d;
  SEXP r   = VECTOR_ELT(db,0);
  DB*  m   = as_db(r);

  update_key(r,child,set,&k);

  /* Update the database in memory */
  if(value == R_NilValue)
    m->del(m,NULL,(DBT *)&k,0);
  else {
    d.data = REAL(value);
    d.size = length(value)*sizeof(double);
    d.flags= 0;
    s = m->put(m,NULL,&k,&d,LOGICAL(ovwr)[0] ? 0 : DB_NOOVERWRITE);
    if(s != 0) report_key_error(s);
  }
  return db;
}

SEXP BNSL_GetDB(SEXP db,SEXP child,SEXP set) {
  int i,s;
  DBT k,d;
  SEXP res;
  SEXP r = VECTOR_ELT(db,0);
  DB*  m = as_db(r);
  d.flags = 0;
  update_key(r,child,set,&k);
  s = m->get(m,NULL,&k,&d,0); 
  if(s == 0) {
    PROTECT(res = allocVector(REALSXP,d.size/sizeof(double)));
    memcpy(REAL(res),(double*)d.data,d.size);
    UNPROTECT(1);
    return res;
  } else {
    return R_NilValue;
  }
}
