#include "common.h"
#include "bitstring.h"

#define Min(A,B) (((A)<(B))?(A):(B))

/**
 * Approximate transposition distance
 */
SEXP BNSL_OrderDistance(SEXP o1,SEXP o2) {
  int i,l;
  if(length(o1) != length(o2))
    error("length mismatch");
  l=0;
  for(i=0;i<length(o1);i++) 
    if(INTEGER(o1)[i] != INTEGER(o2)[i]) l++;
  return ScalarReal(( (double)l)/2.0);
}

SEXP BNSL_HammOrderDistance(SEXP o1,SEXP o2) {
  char *vmax = vmaxget();
  int  l,i,x,y,n;
  char  *m;
  double dist;
  if(length(o1) != length(o2))
    error("lengths do not match");
  l    = length(o1);
  n    = bitstr_size(l);
  dist = (double)(l-1);
  m    = (char*)R_alloc(sizeof(char),n*n);
  memset(m,0,sizeof(char)*n*n);
  if(INTEGER(o1)[0]   != INTEGER(o2)[0])   dist += 1.0;
  if(INTEGER(o1)[l-1] != INTEGER(o2)[l-1]) dist += 1.0;
  for(i=0;i<l-1;i++) {
    x = INTEGER(o1)[i]-1;
    y = INTEGER(o1)[i+1]-1;
    bit_set(m,(n*y)+x);
  }
  for(i=0;i<l-1;i++) {
    x = INTEGER(o2)[i]-1;
    y = INTEGER(o2)[i+1]-1;
    Rprintf("%d %d %d\n",INTEGER(o2)[i],INTEGER(o2)[i+1],bit_test(m,(n*y)+x));
    if(bit_test(m,(n*y)+x)) dist -= 1.0;
  }
  vmaxset(vmax);
  return ScalarReal(dist/2.0);
}
SEXP BNSL_LevenOrderDistance(SEXP o1,SEXP o2) {
  char *vmax = vmaxget();
  int cost,k,i,j,m,n,*d,distance;
  n=length(o1);
  m=length(o2);
  if(n==0 || m==0)
    error("ordering of length 0");
  d = (int*)R_alloc(sizeof(int),(++n)*(++m));
  for(k=0;k<n;k++) d[k]=0;
  for(k=0;k<m;k++) d[k*n]=0;
  for(i=1;i<n;i++)
    for(j=1;j<m;j++) {
      cost = (INTEGER(o1)[i-1]==INTEGER(o2)[j-1])?0:1;
      //Dynamic programming minimization
      d[j*n+i] = Min(
		     Min(d[(j-1)*n+i],d[j*n+(i-1)])+1,
		     d[(j-1)*n+(i-1)]+cost
		     );

    }
  distance = d[n*m-1];
  vmaxset(vmax);
  return ScalarReal((double)distance);
}
