#include <Rdefines.h>
#include "common.h"
#include "bitstring.h"
#include <db.h>
#include <Rmath.h>
#include "fastdb.h"

extern SEXP Rf_findVar1(SEXP,SEXP,SEXPTYPE,int);

SEXP R_do_new_extptr(SEXP class_def) {
  static SEXP s_virtual=NULL,s_prototype,s_className;
  SEXP e,value;
  if(!s_virtual) {
    s_virtual  = install("virtual");
    s_prototype= install("prototype");
    s_className= install("className");
  }
  if(!class_def)
    error("C level NEW macro called with null class definition pointer");
  e = GET_SLOT(class_def,s_virtual);
  if(asLogical(e) != 0) {
    e = GET_SLOT(class_def,s_className);
    error("trying to generate an object in C from a virtual class (\"%s\")",CHAR(asChar(e)));
  }
  e = GET_SLOT(class_def,s_className);
  value = GET_SLOT(class_def,s_prototype);
  if(TYPEOF(value) != EXTPTRSXP)
    error("trying to use specialized externalptr code on non extptr class");

  value = R_MakeExternalPtr(R_ExternalPtrAddr(value),
			    R_ExternalPtrTag(value),
			    R_ExternalPtrProtected(value));
  setAttrib(value,R_ClassSymbol,e);
  return value;
}
void BNSL_FastDBFinalize(SEXP d) {
  DB *db = R_ExternalPtrAddr(d);
  if(db == NULL) return;
  db->close(db,0);
}

SEXP BNSL_MakeFastDB(SEXP file,SEXP vars,SEXP cache) {
  DB *db;
  SEXP d,x;
  if(file != R_NilValue && TYPEOF(file) != STRSXP) error("File must be NULL or a string not %d",TYPEOF(file));
  db_create(&db,NULL,0);
  if(file != R_NilValue)
    db->set_cachesize(db,INTEGER(cache)[0],INTEGER(cache)[1],INTEGER(cache)[2]);
  db->open(db,NULL,
	   file==R_NilValue?NULL:CHAR(STRING_ELT(file,0)),
	   NULL,DB_BTREE,DB_CREATE,0);
  PROTECT(x = R_MakeExternalPtr(db,R_NilValue,R_NilValue));
  R_RegisterCFinalizerEx(x,BNSL_FastDBFinalize,TRUE);
  PROTECT(d = allocVector(VECSXP,1));
  SET_VECTOR_ELT(d,0,x);
  setAttrib(d,install("class"),ScalarString(mkChar("fastdb")));
  setAttrib(d,install("vars"),duplicate(vars));
  UNPROTECT(2);
  return d;
}

SEXP BNSL_MakePsetDB(SEXP _DB) {
  DB *current;
  DB *db;
  SEXP x,d;
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  PROTECT(x = R_MakeExternalPtr(db,R_NilValue,R_NilValue));
  R_RegisterCFinalizerEx(x,BNSL_FastDBFinalize,TRUE);
  PROTECT(d = allocVector(VECSXP,1));
  SET_VECTOR_ELT(d,0,x);
  setAttrib(d,install("class"),ScalarString(mkChar("psetdb")));
  setAttrib(d,install("vars"),duplicate(getAttrib(_DB,install("vars"))));
  DBC   *cursor;
  int   ret,n    = INTEGER(getAttrib(_DB,install("vars")))[0];
  char *pset     = (char*)R_alloc(sizeof(char),bitstr_size(n));
  DBT   key,key2,data;

  current = R_ExternalPtrAddr(VECTOR_ELT(_DB,0));
  current->cursor(current,NULL,&cursor,0);
  memset(&key,0,sizeof(DBT));
  memset(&key2,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));

  key2.data = pset;
  key2.size = sizeof(char)*bitstr_size(n);
  
  memset(pset,0,sizeof(char)*bitstr_size(n));
  for(ret = cursor->c_get(cursor,&key,&data,DB_FIRST);
      !ret;
      ret = cursor->c_get(cursor,&key,&data,DB_NEXT)) {
    int  i,k = key.size/sizeof(int);
    int *P = (int*)key.data;
    memset(pset,0,sizeof(char)*bitstr_size(n));
    for(i=0;i<k;i++) bit_set(pset,P[i]-1);
    db->put(db,NULL,&key2,&data,0);
  }
  UNPROTECT(2);
  return d;
}

SEXP BNSL_FastDBSync(SEXP db) {
  DB *d = (DB*)R_ExternalPtrAddr(VECTOR_ELT(db,0));
  d->sync(d,0);
  return R_NilValue;
}

SEXP BNSL_FastDBGet(SEXP db,SEXP pset) {
  DB *d = (DB*)R_ExternalPtrAddr(VECTOR_ELT(db,0));
  DBT key,data;
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  key.data = INTEGER(pset);
  key.size = pset != R_NilValue?sizeof(int)*length(pset):0;
  if(d->get(d,NULL,&key,&data,0) != 0) 
    return R_NilValue;
  SEXP ret;
  PROTECT(ret = allocVector(REALSXP,data.size/sizeof(double)));
  memcpy(REAL(ret),data.data,data.size);
  UNPROTECT(1);
  return ret;
}

SEXP BNSL_FastDBSet(SEXP db,SEXP pset,SEXP score) {
  DB  *d = (DB*)R_ExternalPtrAddr(VECTOR_ELT(db,0));
  DBT key,data;

  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  if(pset == R_NilValue || length(pset) == 0) {
    int k = 0;
    key.data = &k;
    key.size = sizeof(int);
  } else {
    key.data  = INTEGER(pset);
    key.size  = sizeof(int)*length(pset);
  }
  data.data = REAL(score);
  data.size = sizeof(double)*length(score);
  if(d->put(d,NULL,&key,&data,0) != 0)
    error("database write error");
  return db;
}

fast_score_ctx fast_ctx(SEXP D,int K) {
  SEXP check;
  fast_score_ctx ctx = (fast_score_ctx)R_alloc(sizeof(struct fast_score_ctx),1);
  memset(ctx,0,sizeof(struct fast_score_ctx));
  ctx->d = R_ExternalPtrAddr(VECTOR_ELT(D,0));
  ctx->n = INTEGER(getAttrib(D,install("vars")))[0];
  ctx->K = K;
  ctx->b_vec = (double*)R_alloc(sizeof(double),ctx->n);
  ctx->k_mat = (double*)R_alloc(sizeof(double),ctx->n*(ctx->K+1));

  check = Rf_findVar1(install(".Gamma"),R_GlobalEnv,REALSXP,FALSE);
  ctx->gamma = (check == R_UnboundValue || length(check)==0)?0.0:REAL(check)[0];
  check = Rf_findVar1(install(".Beta"),R_GlobalEnv,REALSXP,FALSE);
  ctx->beta  = (check == R_UnboundValue || length(check)==0)?1.0:REAL(check)[0];
  if(isnan(ctx->beta)) ctx->beta = 0.5;

  //ctx->log_bratio = log(ctx->beta/(1.0-ctx->beta));
  //  ctx->log_opp_b  = log(1.0-ctx->beta);
  
  
  ctx->inv   = (int*)R_alloc(sizeof(int),2*ctx->n);
  ctx->score = (double*)R_alloc(sizeof(int),2*ctx->n);
  ctx->d->cursor(ctx->d,NULL,&(ctx->cur),0);
  return ctx;
}

parent_sets fast_ctx_sets(fast_score_ctx ctx) {
  parent_sets p = (parent_sets)R_alloc(sizeof(struct parent_sets),1);
  int       ret;
  db_create(&(p->db),NULL,0);
  p->db->open(p->db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  memset(&(p->key),0,sizeof(DBT));
  memset(&(p->data),0,sizeof(DBT));
  p->node = (char*)R_alloc(sizeof(char),1+bitstr_size(ctx->n));
  p->pset = p->node + 1;
  p->key.size = 1+bitstr_size(ctx->n);
  p->key.data = p->pset;
  p->data.size = sizeof(double);
  for(ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_FIRST);
      ret == 0;
      ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_NEXT)) {
    int     k = ctx->key.size/sizeof(int);
    int    *P = (int*)(ctx->key.data);
    double *s = (double*)(ctx->data.data);
    //    double  b2 = ctx->log_bratio*((double)k);
    int     m = ctx->data.size/sizeof(double);
    int     i;
    memset(p->pset,0,sizeof(char)*bitstr_size(ctx->n));
    for(i=0;i<k;i++) if(P[i]>0) bit_set(p->pset,P[i]-1);
    for(i=0;i<ctx->n;i++) {
      p->node[0] = i;
      if(s[i] != R_NegInf) {
	p->data.data = &s[i];
	p->db->put(p->db,NULL,&(p->key),&(p->data),0);
      }
    }
  }
  return p;
}

double fast_ctx_score(fast_score_ctx ctx,int *o,int n) {
  int ret;
  int i,j;
  for(i=0;i<n;i++) {
    ctx->inv[o[i]-1] = i;
    ctx->score[i] = R_NegInf;
  }

  for(ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_FIRST);
      ret == 0;
      ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_NEXT)) {
    int     k = ctx->key.size/sizeof(int);
    int    *p = (int*)(ctx->key.data);
    double *s = (double*)(ctx->data.data);
    double  b2 = -ctx->gamma*((double)k);
    int     m = ctx->data.size/sizeof(double);
    if(k == 0 || k > ctx->K) continue;
    for(i=0;i<n;i++) ctx->score;
    if(p[0] == 0)
      for(i=0;i<n;i++) {
	//	b1 = ctx->gamma*ctx->log_opp_b*
	//	  ((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
	ctx->score[o[i]-1] = logspace_add(ctx->score[o[i]-1],s[o[i]-1]);
      }
    else {
      i=0;
      for(j=0;j<k;j++) if(i <= ctx->inv[p[j]-1]) i = ctx->inv[p[j]-1]+1;
      for(;i<n;i++) {
	//	b1 = ctx->log_opp_b*
	//	  ((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
	ctx->score[o[i]-1] = logspace_add(ctx->score[o[i]-1],
					  s[o[i]-1]+b2);
      }
    }
  }
  double r = 0.0;
  for(i=0;i<n;i++)
    r += ctx->score[i];
  return r;
}




/* Scores using the beta vector rather than a global beta */
double fast_ctx_score_bvec(fast_score_ctx ctx,int *o,int n) {
  int ret;
  int i,j;
  
  for(i=0;i<n;i++) {
    ctx->inv[o[i]-1] = i;
    ctx->score[i] = R_NegInf;
  }

  for(ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_FIRST);
      ret == 0;
      ret = ctx->cur->c_get(ctx->cur,&(ctx->key),&(ctx->data),DB_NEXT)) {
    int     k = ctx->key.size/sizeof(int);
    int    *p = (int*)(ctx->key.data);
    double *s = (double*)(ctx->data.data);
    double  b2 = -ctx->gamma*((double)k);
    int     m = ctx->data.size/sizeof(double);
    if(k == 0 || k > ctx->K) continue;
    for(i=0;i<n;i++) ctx->score;
    if(p[0] == 0)
      for(i=0;i<n;i++) {
	//	b1 = ctx->gamma*ctx->log_opp_b*
	//	  ((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
	ctx->score[o[i]-1] = logspace_add(ctx->score[o[i]-1],s[o[i]-1]);
      }
    else {
      i=0;
      for(j=0;j<k;j++) if(i <= ctx->inv[p[j]-1]) i = ctx->inv[p[j]-1]+1;
      for(;i<n;i++) {
	//	b1 = ctx->log_opp_b*
	//	  ((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
	ctx->score[o[i]-1] = logspace_add(ctx->score[o[i]-1],
					  s[o[i]-1]+b2);
      }
    }
  }
  double r = 0.0;
  for(i=0;i<n;i++)
    r += ctx->score[i];
  return r;
}



double bnsl_fast_score_order(DB*d,int*order,int n,int K) {
  char   *vmax = vmaxget();
  int    *inv  = (int*)R_alloc(sizeof(int),n);
  double *score= (double*)R_alloc(sizeof(double),n);
  int     i,j;
  DBT     key,data;
  DBC    *cur;
  double  ret  = 0.0;

  for(i=0;i<n;i++) {
    inv[order[i]-1] = i;
    score[i]        = R_NegInf;
  }
  d->cursor(d,NULL,&cur,0);
  memset(&key,0,sizeof(DBT));memset(&data,0,sizeof(DBT));
  while(cur->c_get(cur,&key,&data,DB_NEXT)==0) {
    int     k = key.size/sizeof(int);
    int    *p = (int*)key.data;
    double *s = (double*)data.data;
    int     m = data.size/sizeof(double);
    if(k == 0 || k > K) continue;
    if(p[0] == 0) 
      for(i=0;i<n;i++) score[order[i]-1] = logspace_add(score[order[i]-1],
							s[order[i]-1]);
    else {
      i=0;
      for(j=0;j<k;j++) if(i <= inv[p[j]-1]) i = inv[p[j]-1]+1;
      for(;i<n;i++)
	score[order[i]-1] = logspace_add(score[order[i]-1],s[order[i]-1]);
    }
  }
  cur->c_close(cur);
  for(i=0;i<n;i++) ret += score[i];
  vmaxset(vmax);
  return ret;
}


SEXP BNSL_FastDBScoreOrder(SEXP db,SEXP order,SEXP K) {
  return ScalarReal(bnsl_fast_score_order((DB*)R_ExternalPtrAddr(VECTOR_ELT(db,0)),
					  INTEGER(order),
					  length(order),
					  INTEGER(K)[0]));
}

struct daginfo {
  double score;
  int    count;
};


int bnsl_fast_sample(DB*d,int*order,int n,int K,DB*dagdb,int dags) {
  char   *vmax = vmaxget();
  int    *inv  = (int*)R_alloc(sizeof(int),n);
  double *score= (double*)R_alloc(sizeof(double),n);
  double *cdf  = (double*)R_alloc(sizeof(double),n);
  double *u    = (double*)R_alloc(sizeof(double),n);
  int     i,j,unique=0;
  DBT     key,data,k2,d2;
  DBC    *cur;

  /* First we need to calculate the normalizing constants */
  for(i=0;i<n;i++) {
    inv[order[i]-1] = i;
    score[i]        = R_NegInf;
  }
  d->cursor(d,NULL,&cur,0);
  memset(&key,0,sizeof(DBT));memset(&data,0,sizeof(DBT));
  while(cur->c_get(cur,&key,&data,DB_NEXT)==0) {
    int     k = key.size/sizeof(int);
    int    *p = (int*)key.data;
    double *s = (double*)data.data;
    double  b1,b2;// = ctx->log_bratio*((double)k);
    int     m = data.size/sizeof(double);
    if(k == 0 || k > K) continue;
    if(p[0] == 0) {
      //      b1 = ctx->gamma*ctx->log_opp_b*
      //	((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
      for(i=0;i<n;i++) score[order[i]-1] = logspace_add(score[order[i]-1],
							s[order[i]-1]);
    } else {
      i=0;
      for(j=0;j<k;j++) if(i <= inv[p[j]-1]) i = inv[p[j]-1]+1;
      for(;i<n;i++) {
	//	b1 = ctx->log_opp_b*
	//	  ((ctx->inv[o[i]-1]<ctx->K)?ctx->inv[o[i]-1]:ctx->K);
	score[order[i]-1] = logspace_add(score[order[i]-1],s[order[i]-1]);
      }
    }
  }
  

  char *dag = (char*)R_alloc(sizeof(char),n*bitstr_size(n));
  struct daginfo di;
  while(dags--) {
    int min,togo = n; 

    memset(dag,0,sizeof(char)*n*bitstr_size(n));
    di.score = 0.0;
    di.count = 1;
    
    /* Sample a set of uniform random values for roulette sampling */
    for(i=0;i<n;i++) {
      cdf[i] = R_NegInf;u[i]=unif_rand();
      /*  Rprintf("%d ",order[i]);*/
    }
    /*Rprintf("\n");*/
    
    int ret  = cur->c_get(cur,&key,&data,DB_FIRST);
    while(ret==0) {
      int  k    = key.size/sizeof(int);
      int *p    = (int*)key.data;
      double *s = (double*)data.data;

      if(k <= K) {
	/* Go to our starting location */
	min=0;
	for(j=0;j<k;j++) if(p[j] && min <= inv[p[j]-1]) min = inv[p[j]-1]+1;
	for(i=min;i<n;i++) {
	  if(u[order[i]-1] <= 1.0) {
	    cdf[order[i]-1] = logspace_add(s[order[i]-1],cdf[order[i]-1]);
	    /* Excellent, we get a roulette match */
	    if(u[order[i]-1] < exp(cdf[order[i]-1]-score[order[i]-1])) {
	      u[order[i]-1] = 2.0;	/* Make it impossible to match again */
	      for(j=0;j<k;j++) bit_set(&dag[(order[i]-1)*bitstr_size(n)],p[j]-1);
	      di.score += s[order[i]-1]-score[order[i]-1];
	      togo--;
	    }
	  }
	}
	if(togo==0) break;
	/*	for(i=0;i<n;i++) 
	  Rprintf("%.2f ",i<min?R_NegInf:exp(cdf[order[i]-1]-score[order[i]-1]));
	Rprintf("[");
	for(i=0;i<k;i++) Rprintf("%d ",p[i]);
	Rprintf("] (%d)\n",togo);*/
      }

      ret = cur->c_get(cur,&key,&data,DB_NEXT);
    }
    if(togo) {
      int k;
      for(i=0;i<n;i++)
	Rprintf("%d (%.2f) ",order[i],u[order[i]-1]);
      Rprintf("\n");
      for(i=0;i<n;i++) 
	Rprintf("%.2f ",exp(cdf[order[i]-1]-score[order[i]-1]));
      Rprintf("(%d %d)\n",togo,dags);
      error("Error sampling dag, still have elements to sample but no more parents!");
    }
    k2.data = dag;
    k2.size = n*bitstr_size(n);
    d2.data = &di;
    d2.size = sizeof(struct daginfo);
    if(dagdb->get(dagdb,NULL,&k2,&d2,0)==0) {
      ((struct daginfo*)d2.data)->count++;
    } else unique++;
    dagdb->put(dagdb,NULL,&k2,&d2,0);
  }
  cur->c_close(cur);
  vmaxset(vmax);
  return unique;
}



extern int IsConsistent(int n,char*rep,int*order,char*topo);

SEXP BNSL_FastDagSampleOld(SEXP _db,SEXP os,SEXP K,SEXP mult) {
  DB   *db;
  DBC  *cur;
  DBT   key,data;
  SEXP dags,sampled,overlap,scores;
  SEXP orders,counts,ret = R_NilValue;
  int  i,k,unique=0,n=0;
  char *vmax = vmaxget();
  char *check;
  GetRNGstate();
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_BTREE,DB_CREATE,0);
  
  orders = GET_SLOT(os,install("orders"));
  counts = GET_SLOT(os,install("counts"));
  n = length(VECTOR_ELT(orders,0));
  check = (char*)R_alloc(sizeof(char),bitstr_size(n));
  for(i=0;i<length(orders);i++) {
    SEXP order = VECTOR_ELT(orders,i);
    int  count = INTEGER(counts)[i]*INTEGER(mult)[0];
    unique += bnsl_fast_sample((DB*)R_ExternalPtrAddr(VECTOR_ELT(_db,0)),
			       INTEGER(order),length(order),
			       INTEGER(K)[0],db,count);
  }
  Rprintf("%d unique DAGs sampled\n",unique);
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("dagSample")));
  PROTECT(scores   = allocVector(REALSXP,unique));
  PROTECT(dags     = allocVector(VECSXP,unique));
  PROTECT(sampled  = allocVector(REALSXP,unique));
  PROTECT(overlap  = allocVector(REALSXP,unique));
  i=0;memset(&key,0,sizeof(DBT));memset(&data,0,sizeof(DBT));
  db->cursor(db,NULL,&cur,0);
  while(cur->c_get(cur,&key,&data,DB_NEXT)==0) {
    char *r            = (char*)key.data;
    struct daginfo *di = (struct daginfo*)data.data;
    SEXP  raw;
    PROTECT(raw = allocVector(RAWSXP,key.size));
    memcpy(RAW(raw),r,key.size);
    SET_VECTOR_ELT(dags,i,raw);
    UNPROTECT(1);
    REAL(scores)[i] = di->score;
    REAL(sampled)[i]= di->count;
    REAL(overlap)[i]= 0.0;
    for(k=0;k<length(orders);k++)
      if(IsConsistent(n,r,INTEGER(VECTOR_ELT(orders,k)),check))
	REAL(overlap)[i] += 1.0;
    i++;
  }
  SET_SLOT(ret,install("vertices"),ScalarInteger(n));
  SET_SLOT(ret,install("dags"),dags);
  SET_SLOT(ret,install("p"),scores);
  SET_SLOT(ret,install("n"),sampled);
  SET_SLOT(ret,install("o"),overlap);
  cur->c_close(cur);
  PutRNGstate();
  vmaxset(vmax);
  
  UNPROTECT(5);
  return ret;
}
