#include "common.h"
#include <Rdefines.h>
#include <db.h>
#include "scoring.h"
#include "bitstring.h"
#include "crossover.h"

void setDims(SEXP x,double rows,double cols) {
  SEXP d;
  PROTECT(d = allocVector(REALSXP,2));
  REAL(d)[0] = rows;
  REAL(d)[1] = cols;
  setAttrib(x,R_DimSymbol,d);
  UNPROTECT(2);
}

/* Simple uniform sampling */
int unif_int(int n) {
  return ((int)((double)n*unif_rand()));
}

/* Roulette Wheel sampling based on Boltzman Weights */
int roulette_int(int n,double t,double *H,double Z) {
  double cdf = 0.0,ru = unif_rand();
  int a = unif_int(n),b=a;
  do {
    cdf += exp((H[a]/t)-Z);
/*     Rprintf("%.2f <= %.2f %.2f %d %d\n",ru,cdf,exp((H[a]/t)-Z),a,b); */
    if(ru <= cdf) return a;
    a = (a + 1) % n;
  } while(a!=b);
  return a ?  a-1 : n-1;
}



/* Mutation Operator */
double operator_M(score_ctx ctx,int*A,
		double score_0,double tk) {
  int a,b,t;
  double score_1;
  a = ((int)((double)ctx->n*unif_rand()));
  do { b = ((int)((double)ctx->n*unif_rand())); } while(a==b);
  t = A[a];
  A[a]=A[b];
  A[b]=t;
/*   { */
/*     int i; */
/*     for(i=0;i<ctx->n;i++) */
/*       Rprintf("%d ",A[i]); */
/*     Rprintf("\n"); */
/*   } */
  score_1 = bnsl_score_ordering(ctx,A,ctx->n);
  /* Acceptance probability */
  if(unif_rand() > exp((score_1-score_0)/tk)) {
    A[b] = A[a];
    A[a] = t;
    return score_0;
  } else
    return score_1;
}


#define CALC_STATS(I) \
  REAL(mu)[I] = 0.0;		 \
  for(j=0;j<pop_size;j++) {	 \
    REAL(mu)[I] += score_0[j];					\
  }								\
  REAL(mu)[I] /= (double)(pop_size);REAL(sig2)[I] = 0.0;	\
  for(j=0;j<pop_size;j++) {					\
    double x= (score_0[j]-REAL(mu)[I]);				\
    REAL(sig2)[I] += x*x;					\
  }								\
  REAL(sig2)[I] /= (double)(pop_size-1);			\
      
  
/**
 * N = iterations
 * qm= mutation rate
 * T = temperature ladder (and pop size)
 */
SEXP BNSL_EMC(SEXP d,SEXP max_parents,SEXP mode,
	      SEXP N,SEXP qm,SEXP T,SEXP start,SEXP XOVER) {
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  double *score_0,*score_1;
  double *exch;
  double  Z;
  int    *P_0,*P_1;
  int  pop_size = length(T);
  int  xover    = pop_size/5;
  int  i,j,ex;

  int  xcount=0,xaccept=0;

  SEXP klass,result,names,orders,scores;
  SEXP state;
  SEXP accept;

  void (*x_operator)(score_ctx,int*,int*,int*,int*);
  switch(INTEGER(XOVER)[0]) {
  case 1:x_operator = operator_CX;break;
  case 2:x_operator = operator_ISX;break;
  case 3:x_operator = operator_ISXU;break;
  case 4:x_operator = operator_ISXA;break;
  }

  
  GetRNGstate();

  PROTECT(orders = allocVector(VECSXP,INTEGER(N)[0]));
  PROTECT(scores = allocVector(REALSXP,INTEGER(N)[0]*pop_size));
  PROTECT(accept = allocVector(REALSXP,pop_size*pop_size));

  exch = (double*)R_alloc(sizeof(double),pop_size*pop_size);
  memset(exch,0,sizeof(double)*pop_size*pop_size);
  memset(REAL(accept),0,sizeof(double)*pop_size*pop_size);

  if(!xover) xover = 1;

  /* Create an initial population of orderings */
  score_0 =(double*)R_alloc(sizeof(double),pop_size);
  score_1 =(double*)R_alloc(sizeof(double),pop_size);
  P_0 = (int*)R_alloc(sizeof(int),pop_size*ctx->n);
  P_1 = (int*)R_alloc(sizeof(int),pop_size*ctx->n);

  if(start == R_NilValue) {
    for(i=0;i<pop_size;i++) {
      int* o = P_0 + (ctx->n*i);	/* The ordering */
      /* Initialize an ordering */
      for(j=0;j<ctx->n;j++) o[j] = j+1;
      /* Now randomize it */
      for(j=0;j<ctx->n;j++) {
	int t,b;do { b = unif_int(ctx->n); } while(b==j);
	t=o[j];
	o[j] = o[b];
	o[b] = t;
      }
      score_0[i] = bnsl_score_ordering(ctx,o,ctx->n);
    }
  } else {
    /* Copy the starting values into the population */
    for(i=0;i<pop_size;i++) {
      int *o = P_0 + (ctx->n*i);
      memcpy(o,INTEGER(VECTOR_ELT(start,i)),ctx->n*sizeof(int));
      score_0[i] = bnsl_score_ordering(ctx,o,ctx->n);
    }
  }



  for(i=0;i<INTEGER(N)[0];i++) {
    /* Update the Boltzman distribution */
    Z = boltz_Z(score_0,1.0,pop_size);
    if(unif_rand() < REAL(qm)[0]) {
      int*o = P_0;
      Rprintf("-");
      /* Mutate the population at random */
      for(j=0;j<pop_size;j++) {
	score_0[j] = operator_M(ctx,o,score_0[j],REAL(T)[j]);
	o += ctx->n;
      }
    } else {
      Rprintf("+");
      /* Select a pair to cross */
      for(j=0;j<xover;j++) {
	double rc,Tyx,Txy,Z2;
	int a = roulette_int(pop_size,1.0,score_0,Z);
	int b;do { b = unif_int(pop_size); } while(a==b);

	xcount++;
	Txy = exp(score_0[a]-Z);
	x_operator(ctx,
		   P_0+a*ctx->n,P_0+b*ctx->n,
		   P_1+a*ctx->n,P_1+b*ctx->n);
	memcpy(score_1,score_0,sizeof(double)*pop_size);
	score_1[a] = bnsl_score_ordering(ctx,P_1+a*ctx->n,ctx->n);
	score_1[b] = bnsl_score_ordering(ctx,P_1+b*ctx->n,ctx->n);
	Z2 = boltz_Z(score_1,1.0,pop_size);
	Tyx = exp(score_1[a]-Z2);
	rc = ((score_1[a]-score_0[a])/REAL(T)[a]) +
	  ((score_1[b]-score_0[b])/REAL(T)[b]);
/* 	Rprintf("%d,%d (%.2f %.2f) %.2f\n",a,b, */
/* 		(score_1[a]-score_0[a])/REAL(T)[a], */
/* 		(score_1[b]-score_0[b])/REAL(T)[b], */
/* 		exp(rc)); */
	if(unif_rand() < exp(rc)*(Tyx/Txy)) {
	  xaccept++;
	  memcpy(P_1+a*ctx->n,P_0+a*ctx->n,sizeof(int)*ctx->n);
	  memcpy(P_1+b*ctx->n,P_0+b*ctx->n,sizeof(int)*ctx->n);
	  score_0[a] = score_1[a];
	  score_0[b] = score_1[b];
	}
      }
    }
    ex=0;
    /* Parallel tempering step */
    for(j=0;j<pop_size;j++) {
      double re,tdiff;
      int b;      
      int a = unif_int(pop_size);
      if(a==0) {
	b=1;
      } else if(a+1 == pop_size) {
	b=a-1;
      } else {
	b = (unif_rand()<=0.5)?a-1:a+1;
      }
      tdiff = (1.0/REAL(T)[b])-(1.0/REAL(T)[a]);
      re    = (score_0[a]-score_0[b])*tdiff;
      exch[b+(a*pop_size)] += 1.0;
      exch[a+(b*pop_size)] += 1.0;
      if(unif_rand() < exp(re)) {
	REAL(accept)[b+(a*pop_size)] += 1.0;
	REAL(accept)[a+(b*pop_size)] += 1.0;
	memcpy(P_1+a*ctx->n,P_0+a*ctx->n,sizeof(int)*ctx->n);
	memcpy(P_0+a*ctx->n,P_0+b*ctx->n,sizeof(int)*ctx->n);
	memcpy(P_0+b*ctx->n,P_1+a*ctx->n,sizeof(int)*ctx->n);
	score_1[a] = score_0[a];
	score_0[a] = score_0[b];
	score_0[b] = score_1[a];
      }
    }
    
    /* Record this iteration */
    {
      SEXP o;
      PROTECT(o = allocVector(INTSXP,ctx->n));
      memcpy(INTEGER(o),P_0,sizeof(int)*ctx->n);
      SET_VECTOR_ELT(orders,i,o);
      memcpy(REAL(scores)+(i*pop_size),score_0,pop_size*sizeof(double));
      UNPROTECT(1);
    }
  }

  for(i=0;i<pop_size*pop_size;i++) {
    if(exch[i] > 0.0)
      REAL(accept)[i] /= exch[i];
  }
  setDims(accept,pop_size,pop_size);
  setDims(scores,pop_size,INTEGER(N)[0]);

  PROTECT(state = allocVector(VECSXP,pop_size));
  for(i=0;i<pop_size;i++) {
    int*o = P_0 + (ctx->n*i);
    SEXP s;
    PROTECT(s = allocVector(INTSXP,ctx->n));
    memcpy(INTEGER(s),o,ctx->n*sizeof(int));
    SET_VECTOR_ELT(state,i,s);
    UNPROTECT(1);
  }

  Rprintf("\n");
  PutRNGstate();
  klass = MAKE_CLASS("order.pop");
  PROTECT(result= NEW_OBJECT(klass));
  SET_SLOT(result,install("orders"),orders);
  SET_SLOT(result,install("qM"),duplicate(qm));
  SET_SLOT(result,install("max.parents"),duplicate(max_parents));
  SET_SLOT(result,install("scores"),scores);
  SET_SLOT(result,install("xover"),duplicate(XOVER));
  SET_SLOT(result,install("xover.accept"),
	   xcount ?
	   ScalarReal((double)xaccept/(double)xcount):ScalarReal(0.0));
  SET_SLOT(result,install("accept.mat"),accept);
  SET_SLOT(result,install("temp.ladder"),duplicate(T));
  SET_SLOT(result,install("state"),state);
  bnsl_score_ctx_done(ctx);
  UNPROTECT(3);
  return result;
}
