#include <R_ext/RS.h>
#include <Rinternals.h>
#include <Rmath.h>
#include <sys/types.h>
#include <limits.h>
#include <db.h>
#include <fenv.h>
#include "bitstring.h"

extern double score_extension(DB*db,DBT*key,DBT*data,
			      int*order,int o_n,
			      int extension,int max_parents,
			      double*scores,int score_max,int n);
double score_extension_max(DB*db,DBT*key,DBT*data,
			   int*order,int o_n,
			   int extension,int max_parents,int n);

double metropolis_moves(DB*db,DBT*key,DBT*data,
			int *o,int n,int max_parents,
			double *scores,int score_max,
			int skip,int*accepted,int use_max) {
  double *o_s,*o_2,Pcurrent,Pnew;
  int i,j,k;

  o_s = (double*)R_alloc(n,sizeof(double));
  o_2 = (double*)R_alloc(n,sizeof(double));
  Pcurrent = 0.0;
  for(i=0;i<n;i++) {
    o_s[i] = (use_max) ?
      score_extension_max(db,key,data,o,i,o[i],max_parents,n)
      : score_extension(db,key,data,o,i,
			o[i],max_parents,
			scores,score_max,n);
    Pcurrent += o_s[i];
  }
  /*  Rprintf("%.2f\n",Pcurrent);*/
  (*accepted) = 0;
  for(i=0;i<skip;i++) {
    int a,b,t;

    /*    for(j=0;j<n;j++)
      printf("%d ",o[j]);
      printf("\n");*/

    memcpy(o_2,o_s,n*sizeof(double));
    a = (int)((double)n*unif_rand());
    do { b = (int)((double)n*unif_rand()); } while(a == b);

    t = o[a];
    o[a] = o[b];
    o[b] = t;
    
    Pnew = Pcurrent;
    for(j=(a < b) ? a : b;j<n;j++) {
      Pnew  -= o_s[j];
      o_s[j] = (use_max) ?
	score_extension_max(db,key,data,o,j,o[j],max_parents,n)
	: score_extension(db,key,data,o,j,
			o[j],max_parents,
			scores,score_max,n);
      Pnew += o_s[j];
    }
    if(log(unif_rand()) > Pnew-Pcurrent) {
      /* Sample was refjected */
      memcpy(o_s,o_2,n*sizeof(double));
      o[b] = o[a];
      o[a] = t;
    } else { 
      if(accepted) (*accepted) = (*accepted) + 1; /* Increase accepted count */
      Pcurrent = Pnew; 
    }
  }
  return Pcurrent;
}

/* Make a metropolis move */
SEXP BNSL_Metropolis(SEXP _db,SEXP order,SEXP M,SEXP iter,SEXP usemax) {
  DB*  db;
  DBT  key,data;
  int  i,n,o_n;
  int     score_max;
  double *scores,t;
  int    *combo,accepted;
  SEXP res,o2;

  GetRNGstate();
  
  get_db_info(_db,&db,(char**)&(key.data),&n);
  key.size   = bitstr_size(n)+1;
  memset((char*)key.data,0,key.size);
  
  score_max = 1;
  for(i=1;i<=INTEGER(M)[0];i++)
    score_max += choose(length(order),i);
  scores = (double*)R_alloc(score_max,sizeof(double));
  
  key.flags  = 0;
  data.flags = 0;
  PROTECT(res = allocVector(VECSXP,3));
  PROTECT(o2  = duplicate(order));
  t =metropolis_moves(db,&key,&data,INTEGER(o2),
		      length(o2),INTEGER(M)[0],
		      scores,score_max,INTEGER(iter)[0],&accepted,LOGICAL(usemax)[0]);
  SET_VECTOR_ELT(res,0,ScalarReal(t));
  SET_VECTOR_ELT(res,1,ScalarReal((double)accepted/(double)INTEGER(iter)[0]));
  SET_VECTOR_ELT(res,2,o2);
  PutRNGstate();
  UNPROTECT(2);
  return res;
}
