#include "common.h"
#include "score_search.h"
#include "scoring.h"

SEXP OrderRandom(int);

void shuffle(int*o,int n) {
  int i,j,t;
  for(i=0;i<n-1;i++) {
    j = i+unif_int(n-i);
    t = o[i];
    o[i] = o[j];
    o[j] = t;  
  }
}

SEXP BNSL_GreedyOptimization(SEXP d,SEXP max_parents,SEXP o,SEXP mode,
			     SEXP time) {
  SEXP      ret;
  char     *vmax = vmaxget();
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  int      *swap= (int*)R_alloc(sizeof(int),ctx->n);
  double score,s1,s2;
  int    steps,x,y,z,t,l = ctx->n;
  PROTECT(ret = (o!=R_NilValue)?duplicate(o):OrderRandom(ctx->n));

  score = -bnsl_score_ordering(ctx,INTEGER(ret),l);
  for(x=0;x<l;x++) swap[x] = x;
  steps = 0;

 next_iter:
  if(steps++ < REAL(time)[0]) {
    shuffle(swap,l);z=0;
    for(x=0;x<l-1;x++) {
      t = INTEGER(ret)[swap[x]];
      for(y=x+1;y<l;y++) {
	s1 = -bnsl_score_ordering_update(ctx,INTEGER(ret),l,x,y);
	INTEGER(ret)[x] = INTEGER(ret)[y];
	INTEGER(ret)[y] = t;
	s2 = -bnsl_score_ordering_update(ctx,INTEGER(ret),l,x,y);
	if(s2 >= s1) {
	  INTEGER(ret)[y] = INTEGER(ret)[x];
	  INTEGER(ret)[x] = t;
	  z++;
	} else { 
	  score = score - s1 + s2;
	  Rprintf("(%d) %.2f...",z,score);
	  goto next_iter; 
	}
      }
    }
  }
  Rprintf("\n");
  setAttrib(ret,install("score"),ScalarReal(score));
  setAttrib(ret,install("steps"),ScalarReal((double)steps));
  vmaxset(vmax);
  UNPROTECT(1);
  return ret;
}


SEXP BNSL_GreedySearch(SEXP d,SEXP order,SEXP max_parents,SEXP mode) {
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  
  SEXP   ret,names;
  int    moved,x,y,z,n;
  double score = bnsl_score_ordering(ctx,INTEGER(order),length(order));
  int*   P = (int*)R_alloc(sizeof(int),length(order)+2);
  n=0;
  do {
    
    double best_score;
    int    best_swap[2];

    int cur_swap[2],temp;

    moved = 0;
    n++;
    inittwiddle(2,length(order),P);
    cur_swap[0] = length(order)-2;
    cur_swap[1] = length(order)-1;
    temp = INTEGER(order)[cur_swap[0]];
    INTEGER(order)[cur_swap[0]] = INTEGER(order)[cur_swap[1]];
    INTEGER(order)[cur_swap[1]] = temp;
    best_score = bnsl_score_ordering(ctx,INTEGER(order),length(order));
    INTEGER(order)[cur_swap[1]] = INTEGER(order)[cur_swap[0]];
    INTEGER(order)[cur_swap[0]] = temp;

    best_swap[0]=cur_swap[0];best_swap[1]=cur_swap[1];
    
    /*    Rprintf("[%d %d %.2f]\n",cur_swap[0],cur_swap[1],best_score);*/
    while(!twiddle(&x,&y,&z,P)) {
      double temp_score;
      
      cur_swap[z] = x;
      /*      Rprintf("%d %d] ",cur_swap[0],cur_swap[1]);*/
      
      /* Test the swap */
      temp = INTEGER(order)[cur_swap[0]];
      INTEGER(order)[cur_swap[0]] = INTEGER(order)[cur_swap[1]];
      INTEGER(order)[cur_swap[1]] = temp;    
      temp_score = bnsl_score_ordering(ctx,INTEGER(order),length(order));
      INTEGER(order)[cur_swap[1]] = INTEGER(order)[cur_swap[0]];
      INTEGER(order)[cur_swap[0]] = temp;
      /*      Rprintf("%.2f\n",temp_score);*/
      if(temp_score > best_score) {
	best_score = temp_score;   
	best_swap[0]=cur_swap[0];best_swap[1]=cur_swap[1];
	}
    }
    if(best_score > score) {
      /*      Rprintf("Updated score: %.2f\n",best_score);*/
      score = best_score;     
      temp = INTEGER(order)[best_swap[0]];
      INTEGER(order)[best_swap[0]] = INTEGER(order)[best_swap[1]];
      INTEGER(order)[best_swap[1]] = temp; 
      moved = 1;
    }
  } while(moved);

  PROTECT(ret = allocVector(VECSXP,3));
  PROTECT(names = allocVector(STRSXP,3));
  SET_STRING_ELT(names,0,mkChar("ordering"));SET_VECTOR_ELT(ret,0,order);
  SET_STRING_ELT(names,1,mkChar("score"));SET_VECTOR_ELT(ret,1,ScalarReal(score));
  SET_STRING_ELT(names,2,mkChar("iter"));SET_VECTOR_ELT(ret,2,ScalarReal((double)n));
  setAttrib(ret,R_NamesSymbol,names);
  bnsl_score_ctx_done(ctx);
  UNPROTECT(2);
  return ret;
}

SEXP BNSL_SimulatedAnnealing(SEXP d,SEXP order,
			     SEXP max_parents,SEXP mode,
			     SEXP N,SEXP T0,SEXP TN) {

  SEXP temps,scores,ret,names;
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);

  int i,j,l;
  int  x,y,ox,oy;
  double A,Ti = REAL(T0)[0];
  double score,score_new;
  double Accept;

  PROTECT(temps  = allocVector(REALSXP,INTEGER(N)[0]));
  PROTECT(scores = allocVector(REALSXP,INTEGER(N)[0]));

  A = log(REAL(T0)[0] - REAL(TN)[0])/log((double)INTEGER(N)[0]);
  l = length(order);
  score = bnsl_score_ordering(ctx,INTEGER(order),l);
  for(i=0;i<INTEGER(N)[0];i++) {
    Ti = REAL(T0)[0] - pow(((double)i),A);

    /* Choose a new ordering */
    x  = (int)(unif_rand()*(double)l);y = x;
    while(y==x) { y = (int)(unif_rand()*(double)l); }
    ox = INTEGER(order)[x];
    oy = INTEGER(order)[y];
    /*    Rprintf("(%d %d) ",x,y);*/
    INTEGER(order)[x] = oy;INTEGER(order)[y] = ox;
    /*    for(j=0;j<l;j++) Rprintf("%d ",INTEGER(order)[j]);*/
    score_new = bnsl_score_ordering(ctx,INTEGER(order),l);
    Accept = exp((score_new - score)/Ti);
    /*    Rprintf(">%.2f %.2f %.2f\n",score_new,Accept,Ti); */
    
    if(Accept > 1) Accept = 1;
    if(unif_rand() > Accept) {
      INTEGER(order)[x] = ox;INTEGER(order)[y] = oy;
    } else {
      score = score_new;
    }
    REAL(temps)[i] = Ti;
    REAL(scores)[i] = score;
  }
  
  PROTECT(ret = allocVector(VECSXP,3));
  PROTECT(names = allocVector(STRSXP,3));
  SET_STRING_ELT(names,0,mkChar("ordering"));
  SET_STRING_ELT(names,1,mkChar("score"));
  SET_STRING_ELT(names,2,mkChar("temperature"));
  SET_VECTOR_ELT(ret,0,order);
  SET_VECTOR_ELT(ret,1,scores);
  SET_VECTOR_ELT(ret,2,temps);
  setAttrib(ret,R_NamesSymbol,names);
  bnsl_score_ctx_done(ctx);
  UNPROTECT(4);
  return ret;

}
