#include <R_ext/RS.h>
#include <Rinternals.h>
#include <Rmath.h>
#include <sys/types.h>
#include <limits.h>
#include <db.h>
#include "bitstring.h"
#include "scoring.h"

/*
SEXP BNSL_EdgeProb(SEXP d,SEXP max_parents,SEXP mode,
		      SEXP orders,SEXP FROM,SEXP TO) {
  score_ctx ctx = bnsl_score_ctx(d,INTEGER(max_parents)[0],INTEGER(mode)[0]);
  SEXP      out;
  int       o;
  PROTECT(out = allocVector(REALSXP,length(orders)));
  for(o=0;o<length(orders);o++) {
    int* order = INTEGER(VECTOR_ELT(orders,o));
    int  i,from=0,to=0;
    for(i=0;i<ctx->n;i++) {
      if(order[i]==INTEGER(FROM)[0]) from = i;
      if(order[i]==INTEGER(TO)[0]) to = i;
    }
    if(to >= from) 
      REAL(out)[o] = R_NegInf;
    else
      REAL(out)[o] = bnsl_score_feature(ctx,order,to,
					 INTEGER(FROM)[0],
					 INTEGER(TO)[0]);
  }
  UNPROTECT(1);
  return out;
}
*/
int findNode(SEXP names,SEXP e) {
  int i;
  for(i=0;i<length(names);i++) {
    if(strcmp(CHAR(STRING_ELT(names,i)),CHAR(e)) == 0) return i;
  }
  return -1;
}

SEXP BNSL_TopologicalSort(SEXP edgeL,SEXP names,SEXP n) {
  SEXP output;
  int* counts,*working,i,j,k,l,m;
  
  PROTECT(output = allocVector(VECSXP,INTEGER(n)[0]));


  /* Build Counts List */
  l = length(edgeL);
  
  counts = (int*)R_alloc(l,sizeof(int));
  working= (int*)R_alloc(l,sizeof(int));
  memset(counts,0,sizeof(int)*l);
  for(i=0;i<l;i++) {
    SEXP edges = VECTOR_ELT(edgeL,i);
    for(j=0;j<length(edges);j++) {
      counts[findNode(names,STRING_ELT(edges,j))]++;
    }
  }

  GetRNGstate();
  /* Now reconstruct */
  for(i=0;i<INTEGER(n)[0];i++) {
    SEXP   order,edges;
    double pdf,ru,cdf;
    
    PROTECT(order = allocVector(INTSXP,l));
    memcpy(working,counts,sizeof(int)*l);
    for(j=0;j<l;j++) {
      pdf = 0.0;
      for(k=0;k<l;k++) if(working[k] == 0) pdf += 1.0;
      pdf = 1/pdf;cdf = 0.0;ru = unif_rand();
      for(k=0;k<l;k++) {
	if(working[k] == 0) cdf += pdf;
	if(ru < cdf) break;
      }
      edges = VECTOR_ELT(edgeL,k);
      for(m=0;m<length(edges);m++)
	working[findNode(names,STRING_ELT(edges,m))]--;
      working[k]--;  

      INTEGER(order)[j] = k+1;
    }
    if(INTEGER(n)[0] == 1) {
      UNPROTECT(2);
      return order;
    }
    SET_VECTOR_ELT(output,i,order);
    UNPROTECT(1);
  }
  PutRNGstate();
  UNPROTECT(1);
  return output;
}
