#include <bitstring.h>
#include "common.h"
#include <Rdefines.h>
#include <Rmath.h>
#include <db.h>
#include <time.h>
#include "scoring.h"


typedef struct _daginfo {
  double score;
  int    sampled;
} daginfo;

double GraphDist(int n,Rbyte *a,Rbyte *b) {
  int i,j,k,size = bitstr_size(n);
  char *r;
  double dist = 0;
  for(i=0;i<n;i++) {
    for(j=0;j<n;j++) {
      /* Missing or reversal */
      if(bit_test(a+i*size,j) != bit_test(b+i*size,j)) dist += 1.0;
    }
  }
  return dist;
}

SEXP BNSL_DistMatrix(SEXP n,SEXP ar,SEXP br) {
  SEXP ret;
  SEXP dims;
  int  i,j,t;
  PROTECT(ret = allocVector(REALSXP,length(ar)*length(br)));
  PROTECT(dims= allocVector(INTSXP,2));
  INTEGER(dims)[0] = length(br);
  INTEGER(dims)[1] = length(ar);
  t =0;
  for(i=0;i<length(ar);i++) {
    for(j=0;j<length(br);j++) {
      REAL(ret)[t++] = GraphDist(INTEGER(n)[0],
				 RAW(VECTOR_ELT(ar,i)),
				 RAW(VECTOR_ELT(br,i)));
    }
  }
  setAttrib(ret,R_DimSymbol,dims);
  UNPROTECT(1);
  
  return ret;
}

#define WHERE(I) ctx[I]->scores[pos[I]]
#define NORM(I)  ctx[i]->norm
#define NEXT(I)  ctx[I]->scores[pos[I]+1]

double CalcPDF(pset_ctx* ctx,int*pos,int n) {
  int i;
  double pdf = 0.0,norm = 0.0;
  for(i=0;i<n;i++) {
    pdf += WHERE(i);norm += NORM(i);
  }
  return exp(pdf-norm);
}

void rep2dag(char*rep,dag_ctx dag) {
  int i,j;
  bnsl_dag_clear(dag);
  for(i=0;i<dag->n;i++) {
    for(j=0;j<dag->n;j++) 
      if(bit_test(rep,j)) {
	bnsl_dag_set_edge(dag,j+1,i+1,1);
      }
    rep += bitstr_size(dag->n);
  }
}

void dag2rep(dag_ctx dag,char *rep) {
  int i;
  for(i=0;i<dag->n;i++) {
    memcpy(rep,bnsl_dag_parent_set(dag,i+1),bitstr_size(dag->n));
    rep += bitstr_size(dag->n);
  }
}

double ImportanceSampleDag(pset_ctx *ctx,int n,char*rep) {
  int    i,j;
  double ru,cdf,score = 0;
  int     s = bitstr_size(n);
  char  *pa,*ra=rep;
  for(i=0;i<n;i++) {
    ra = rep + ((ctx[i]->v-1)*s);
    if(ctx[i]->used > 0) {
      ru = unif_rand();cdf = 0.0;
      for(j=0;j<ctx[i]->used-1;j++) {
	/*	Rprintf(" %.2f ",exp(ctx[i]->scores[j]-ctx[i]->norm),
		ctx[i]->order[j],ctx[i]->v);*/
	cdf += exp(ctx[i]->scores[j]-ctx[i]->norm);
	if(ru <= cdf) break;
      }
      pa = ctx[i]->pset+(s*ctx[i]->order[j]);
      /*for(j=0;j<ctx[i]->n;j++)
	if(bit_test(pa,j)) Rprintf("%d ",j+1);
            Rprintf("->%d \n",ctx[i]->v);*/

      memcpy(ra,pa,s);
      score += ctx[i]->scores[j]-ctx[i]->norm;
    } else
      memset(ra,0,s);
    ra += s;
  }
  /* Return the importance weight */
  return score;
}

double SampleGraph(pset_ctx* ctx,dag_ctx dag,SEXP o,char*rep,
		   int*pos) {
  int    n   = dag->n;
  double ru;
  double cdf,pdf=1.0;
  int    i,j,y;
  char  *x;
  memset(rep,0,bitstr_size(dag->n)*dag->n);
  memset(pos,0,sizeof(int)*n);
  pdf = cdf = CalcPDF(ctx,pos,n);
  for(i=1;i<n;i++) {
    ru = unif_rand();cdf = 0.0;
    for(j=0;j<ctx[i]->used;j++) {
      cdf += exp(ctx[i]->scores[j]-ctx[i]->norm);
      if(ru < cdf) break;
    }
    y = INTEGER(o)[i]-1;
    x = ctx[i]->pset + ctx[i]->order[j]*bitstr_size(n);
    memcpy(rep+y*bitstr_size(n),x,bitstr_size(n));
    pdf *= exp(ctx[i]->scores[j]-ctx[i]->norm);
  }
  rep2dag(rep,dag);
  return pdf;
}

int IsConsistent(int n,char*rep,int*order,char*topo) {
  int   i,j;
  char *p;
  memset(topo,0,bitstr_size(n));
  /*  for(i=0;i<n;i++) Rprintf("%d ",order[i]);
  Rprintf("\n");*/
  for(i=n-1;i>=0;i--) {
    bit_set(topo,order[i]-1);
    p = rep + bitstr_size(n)*(order[i]-1);
    for(j=0;j<n;j++) {
      if(bit_test(p,j)) {
	/*	Rprintf("%d",j+1); */
	if(bit_test(topo,j)) { 
	  /*Rprintf("*->%d\n",order[i]);*/ return 0; 
	}/* else {
	  Rprintf(" ");
	  }*/
      }
    }
    /*    Rprintf("->%d\n",order[i]);*/
  }
  return 1;
}

SEXP BNSL_ConsistentOrders(SEXP DAG,SEXP ORDERS) {
  int   n = sqrt(length(DAG));
  char *r = (char*)R_alloc(sizeof(char),bitstr_size(n)*n);
  char *c = (char*)R_alloc(sizeof(char),n);
  int  i,j;

  double *d = REAL(DAG);
  char   *f = r;
  memset(r,0,bitstr_size(n)*n*sizeof(char));
  for(i=0;i<n;i++) {
    for(j=0;j<n;j++) if(d[j]) bit_set(f,j);
    f += bitstr_size(n);
    d += n;
  }


  SEXP ret;
  PROTECT(ret = allocVector(LGLSXP,length(ORDERS)));
  for(i=0;i<length(ORDERS);i++) {
    LOGICAL(ret)[i] = IsConsistent(n,r,INTEGER(VECTOR_ELT(ORDERS,i)),c);
  }
  UNPROTECT(1);
  return ret;
}

struct setdata {
  unsigned int count;
  double       score;
};

SEXP BNSL_OrderSet(SEXP orders,SEXP scores,SEXP skip,SEXP thin) {
  SEXP ret,counts;
  DB *db;DBT key,data;DBC*cur;
  int i,n,pos,unique;
  struct setdata info,*pinfo;
  
  char *vmax = (char*)vmaxget();
  int  *tmp;
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_HASH,DB_CREATE,0);
  n = length(VECTOR_ELT(orders,0));
  tmp = (int*)R_alloc(sizeof(int),n);
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  key.size = sizeof(int)*n;
  key.data = tmp;
  unique = 0;
  for(i=INTEGER(skip)[0];i<length(orders);i+=INTEGER(thin)[0]) {
    memcpy(tmp,INTEGER(VECTOR_ELT(orders,i)),sizeof(int)*n);
    if(db->get(db,NULL,&key,&data,0) == 0) {
      pinfo = ((struct setdata*)data.data);
      pinfo->count++;
    } else {
      unique++;
      data.data = pinfo = &info;
      data.size = sizeof(struct setdata);
      pinfo->count = 1;
      pinfo->score = (scores != R_NilValue)?REAL(scores)[i]:0.0;
    }
    db->put(db,NULL,&key,&data,0);
  }

  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("orderSet")));
  PROTECT(orders = allocVector(VECSXP,unique));
  PROTECT(scores = allocVector(REALSXP,unique));
  PROTECT(counts = allocVector(INTSXP,unique));
  db->cursor(db,NULL,&cur,0);
  pos = 0;
  while(cur->c_get(cur,&key,&data,DB_NEXT)==0) {
    SEXP o;
    PROTECT(o = allocVector(INTSXP,n));
    memcpy(INTEGER(o),(int*)(key.data),n*sizeof(int));
    SET_VECTOR_ELT(orders,pos,o);
    pinfo = (struct setdata*)(data.data);
    INTEGER(counts)[pos] = pinfo->count;
    REAL(scores)[pos++]  = pinfo->score;
    UNPROTECT(1);
  }
  cur->c_close(cur);
  SET_SLOT(ret,install("orders"),orders);
  SET_SLOT(ret,install("counts"),counts);
  SET_SLOT(ret,install("scores"),scores);
  vmaxset(vmax);
  db->close(db,0);
  UNPROTECT(4);
  return ret;
}

struct odata {
  int x;
  int y;
};

SEXP BNSL_OrderOverlap(SEXP x,SEXP y,SEXP w) {
  SEXP ret;
  DB *db;DBT key,data;DBC*cur;
  int total,unique,i,s,n,shared,width = INTEGER(w)[0];
  struct odata info,*pinfo;

  int MoveOrder(int*order,int x,int add) {
    int shared_orig,shared_now;
    int move;
    key.data = order;
    key.size = s*sizeof(int);
    if(db->get(db,NULL,&key,&data,0) == 0) {
      pinfo = ((struct odata*)data.data);
      shared_orig = (pinfo->x > 0 && pinfo->y > 0)?1:0;
      
      if(x) 
	pinfo->x += (add)?1:-1;
      else
	pinfo->y += (add)?1:-1;
      
      shared_now  = (pinfo->x > 0 && pinfo->y > 0)?1:0;
      
      if(shared_orig && !shared_now)
	move = -1;
      else if(!shared_orig && shared_now)
	move =  1;
      else
	move = 0;
      /* Rprintf("(%d %d %d)",pinfo->x,pinfo->y,move); */
      total++;
    } else {
      pinfo = &info;
      pinfo->x = (x)?1:0;
      pinfo->y = (x)?0:1;
      data.data = pinfo;
      data.size = sizeof(struct odata);
      unique++;total++;
      move = 0;
      /* Rprintf("(unique)"); */
    }
    db->put(db,NULL,&key,&data,0);
    return move;
  }



  
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_HASH,DB_CREATE,0);
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));  
  n = length(x);s = length(VECTOR_ELT(x,0));
  PROTECT(ret = allocVector(INTSXP,length(x)-width));

  total = unique = shared = 0;
  key.size = s*sizeof(int);
  /* Initialize the set */
  for(i=0;i<width;i++) {
    shared += (MoveOrder(INTEGER(VECTOR_ELT(x,i)),1,1)
	       +MoveOrder(INTEGER(VECTOR_ELT(y,i)),0,1));
    /*    Rprintf("[%d]\n",shared); */
  }
  

  do {
    INTEGER(ret)[i-width] = shared;
    shared += (MoveOrder(INTEGER(VECTOR_ELT(x,i-width)),1,0)
	       +MoveOrder(INTEGER(VECTOR_ELT(y,i-width)),0,0));
    shared += (MoveOrder(INTEGER(VECTOR_ELT(x,i)),1,1)
	       +MoveOrder(INTEGER(VECTOR_ELT(y,i)),0,1));
/*    Rprintf("[%d]\n",shared); */
  } while(++i<n);

  Rprintf("%d unique,%d total\n",unique,total);
  
  UNPROTECT(1);
  return ret;
}


typedef struct _move {
  char i,j,rem;
} move;

int  count_moves(dag_ctx dag,int*o,move*movelist) {
  int i,j,t=0;
  for(i=0;i<dag->n;i++)
    for(j=0;j<i;j++) {
      if(bnsl_dag_has_edge(dag,o[j],o[i])) {
	movelist[t].i   = o[i];
	movelist[t].j   = o[j];
	movelist[t].rem = 1;
	t++;
      } else if(bnsl_dag_degree(dag,o[i]) < dag->k) {
	movelist[t].i   = o[i];
	movelist[t].j   = o[j];
	movelist[t].rem = 0;
	t++;
      }
    }
  return t;
}

int StepDag(dag_ctx dag,int*order,move*movelist,double* score,int* t1) {
  double new_score,mh;
  int    t2,j = unif_int(*t1);
  int    from,to,rem;

  from = movelist[j].j;
  to   = movelist[j].i;
  rem  = movelist[j].rem;
  bnsl_dag_set_edge(dag,from,to,rem?0:1);
  t2 = count_moves(dag,order,movelist);
  new_score = -bnsl_dag_score(dag);
  mh = exp((*score)-new_score)*(((double)(*t1))/((double)t2));
  /*  Rprintf("exp(%.2f-%.2f)*(%d/%d)=%.2f\n",*score,new_score,*t1,t2,mh);*/
  if(mh >= 1.0 || unif_rand() < mh) {
    *score = new_score;
    *t1    = t2;
    return 1;
  } else {
    bnsl_dag_set_edge(dag,from,to,rem);
    count_moves(dag,order,movelist);
    return 0;
  }
}

/** Simple Importance Sampling for DAGS */
void ISampleDagsFromOrder(DB*db,dag_ctx dag,int *o,int o_n,int c,int n,
			  pset_ctx *pset,char*rep,int*unique,char*topo) {
  DBT     key,data;
  daginfo info;
  int     i,j,k;
  double  score;
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  key.data = rep;
  key.size = sizeof(char)*dag->n*bitstr_size(dag->n);
  for(i=0;i<c*n;i++) {
    score = ImportanceSampleDag(pset,dag->n,rep);
    /*    score = -bnsl_dag_score(dag);
	  dag2rep(dag,rep);*/
    if(db->get(db,NULL,&key,&data,0) == 0) {
      daginfo *di = (daginfo*)(data.data);
      if(!IsConsistent(o_n,rep,o,topo))
	error("Sampled Inconsistent DAG");

	
      di->sampled++;
      //      di->score = logspace_add(di->score,score);
      db->put(db,NULL,&key,&data,0);
    } else {
      data.data = &info;
      data.size = sizeof(daginfo);
      info.sampled = 1;
      info.score   = score;
      db->put(db,NULL,&key,&data,0);
      (*unique)++;
    }    
  }
}

/** Simple MCMC for Sampling DAGS **/
void SampleDagsFromOrder(DB*db,dag_ctx dag,int *o,int o_n,int c,int n,
			 move*movelist,char*rep,int*unique) {
  DBT     key,data;
  daginfo info;
  int     i,j,k,t1,a,r,x;
  double  score;
  memset(&key,0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  key.data = rep;
  key.size = sizeof(char)*dag->n*bitstr_size(dag->n);
  


  for(k=0;k<c;k++) {
    a     =  r = 0;
    /* Initialize a starting location */
    bnsl_dag_clear(dag);
    for(i=0;i<o_n;i++) {
      int max_k = unif_int(i<dag->k?i+1:dag->k+1);
      if(max_k==0) {
	bnsl_dag_update_score(dag,o[i]);
      } else
	while(max_k>0) {
	  j = unif_int(i);
	  if(!bnsl_dag_has_edge(dag,o[j],o[i])) {
	    bnsl_dag_set_edge(dag,o[j],o[i],1);
	    max_k--;
	  }
	}
    }
    score = -bnsl_dag_score(dag);
    dag2rep(dag,rep);
    t1    =  count_moves(dag,o,movelist);

    x = *unique;
    /* 2N steps, N for burnin */
    for(i=0;i<n*2;i++) {
      if(StepDag(dag,o,movelist,&score,&t1)) {
	dag2rep(dag,rep);
	a++; 
      } else r++;
      if(i>=n) {
	if(db->get(db,NULL,&key,&data,0) == 0) {
	  daginfo *di = (daginfo*)(data.data);
	  di->sampled++;
	  db->put(db,NULL,&key,&data,0);
	} else {
	  data.data = &info;
	  data.size = sizeof(daginfo);
	  info.score   = -score;
	  info.sampled = 1;
	  db->put(db,NULL,&key,&data,0);
	  (*unique)++;
	}
      }
    }
    Rprintf("%d (%d %d %.2f %d) ",k,a,r,((double)a)/((double)a+(double)r),(*unique)-x);
  }
  Rprintf("\n");

}

SEXP BNSL_DagSampleFromOrders(SEXP d,SEXP max_parents,SEXP orderset,SEXP N) {

  SEXP ret = R_NilValue;
  SEXP orders,counts;
  DB*db;
  char *vmax  = vmaxget();
  dag_ctx dag = bnsl_dag_ctx(d,INTEGER(max_parents)[0]);
  char   *rep = (char*)R_alloc(sizeof(char),dag->n*bitstr_size(dag->n));
  move    *ml = (move*)R_alloc(sizeof(move),dag->n*dag->n);
  char  *check= (char*)R_alloc(sizeof(char),bitstr_size(dag->n));
  int i,j,k,unique  = 0;

  
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_HASH,DB_CREATE,0);
  GetRNGstate();
  orders = GET_SLOT(orderset,install("orders"));
  counts = GET_SLOT(orderset,install("counts"));
  for(i=0;i<length(orders);i++) {
    SEXP order = VECTOR_ELT(orders,i);
    Rprintf("%d (%d %d): ",i,INTEGER(counts)[i],INTEGER(N)[0]);
    SampleDagsFromOrder(db,dag,INTEGER(order),dag->n,
			INTEGER(counts)[i],INTEGER(N)[0],ml,rep,&unique);
  }

  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("dagSample")));
  {
    SEXP scores;
    SEXP reps;
    SEXP sampled;
    SEXP overlap;
    DBC* cur;DBT key,data;
    int pos=0;
    
    PROTECT(scores = allocVector(REALSXP,unique));
    PROTECT(reps   = allocVector(VECSXP,unique));
    PROTECT(sampled= allocVector(REALSXP,unique));
    PROTECT(overlap= allocVector(REALSXP,unique));
    db->cursor(db,NULL,&cur,0);
    memset(&key,0,sizeof(DBT));
    memset(&data,0,sizeof(DBT));

    while(cur->c_get(cur,&key,&data,DB_NEXT)==0) {
      char*    r  = (char*)(key.data);
      daginfo *di = (daginfo*)(data.data);
      SEXP     raw;
      PROTECT(raw = allocVector(RAWSXP,dag->n*bitstr_size(dag->n)));
      memcpy(RAW(raw),r,dag->n*bitstr_size(dag->n));
      SET_VECTOR_ELT(reps,pos,raw);
      UNPROTECT(1);
      REAL(scores)[pos]  = di->score;
      REAL(sampled)[pos] = di->sampled;
      REAL(overlap)[pos] = 0.0;
      for(k=0;k<length(orders);k++)
	if(IsConsistent(dag->n,r,INTEGER(VECTOR_ELT(orders,k)),check))
	  REAL(overlap)[pos] += 1.0;
      pos++;
    }
    cur->c_close(cur);
    SET_SLOT(ret,install("vertices"),ScalarInteger(dag->n));
    SET_SLOT(ret,install("dags"),reps);
    SET_SLOT(ret,install("p"),scores);
    SET_SLOT(ret,install("n"),sampled);
    SET_SLOT(ret,install("o"),overlap);
    UNPROTECT(4);
  }
  PutRNGstate();
  db->close(db,0);
  vmaxset(vmax);
  UNPROTECT(1);
  return ret;
}

SEXP BNSL_DagISampleFromOrders(SEXP d,SEXP max_parents,SEXP orderset,SEXP N) {

  SEXP ret = R_NilValue;
  SEXP orders,counts;
  time_t time_start,time_end;
  DB*db;
  char     *vmax = vmaxget();
  dag_ctx   dag  = bnsl_dag_ctx(d,INTEGER(max_parents)[0]);
  pset_ctx *pset = (pset_ctx*)R_alloc(sizeof(pset_ctx),dag->n);
  char     *rep  = (char*)R_alloc(sizeof(char),dag->n*bitstr_size(dag->n));
  char     *check= (char*)R_alloc(sizeof(char),bitstr_size(dag->n));
  int i,j,k,unique  = 0;

  /* Prepare the parent set list for each entry */
  for(i=0;i<dag->n;i++) {
    bnsl_dag_update_score(dag,i+1);
    pset[i] = bnsl_pset_ctx(d,INTEGER(max_parents)[0]);
  }
  
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_HASH,DB_CREATE,0);
  GetRNGstate();
  orders = GET_SLOT(orderset,install("orders"));
  counts = GET_SLOT(orderset,install("counts"));
  for(i=0;i<length(orders);i++) {
    SEXP order = VECTOR_ELT(orders,i);
    for(j=0;j<dag->n;j++) {
      pset[j]->prior.gamma = 0.0; /* We will handle the prior ourselves */
      bnsl_pset_prepare(pset[j],INTEGER(order),j,INTEGER(order)[j]);
    }
    Rprintf("%d (%d %d): ",i,INTEGER(counts)[i],INTEGER(N)[0]);
    time_start = time(NULL);
    ISampleDagsFromOrder(db,dag,INTEGER(order),dag->n,
			 INTEGER(counts)[i],INTEGER(N)[0],pset,rep,&unique,
			 check);
    time_end = time(NULL);
    Rprintf("%d (%.2f seconds)\n",unique,difftime(time_end,time_start));
  }

  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("dagSample")));
  {
    SEXP scores;
    SEXP reps;
    SEXP sampled;
    SEXP overlap;
    DBC* cur;DBT key,data;
    int pos=0;
    
    PROTECT(scores = allocVector(REALSXP,unique));
    PROTECT(reps   = allocVector(VECSXP,unique));
    PROTECT(sampled= allocVector(REALSXP,unique));
    PROTECT(overlap= allocVector(REALSXP,unique));
    db->cursor(db,NULL,&cur,0);
    memset(&key,0,sizeof(DBT));
    memset(&data,0,sizeof(DBT));

    while(cur->c_get(cur,&key,&data,DB_NEXT)==0) {
      char*    r  = (char*)(key.data);
      daginfo *di = (daginfo*)(data.data);
      SEXP     raw;
      PROTECT(raw = allocVector(RAWSXP,dag->n*bitstr_size(dag->n)));
      memcpy(RAW(raw),r,dag->n*bitstr_size(dag->n));
      SET_VECTOR_ELT(reps,pos,raw);
      UNPROTECT(1);
      REAL(scores)[pos]  = di->score;
      REAL(sampled)[pos] = di->sampled;
      REAL(overlap)[pos] = 0.0;
      for(k=0;k<length(orders);k++)
	if(IsConsistent(dag->n,r,INTEGER(VECTOR_ELT(orders,k)),check))
	  REAL(overlap)[pos] += 1; //INTEGER(counts)[k];
      if(REAL(overlap)[pos] == 0.0) {
	char *q = r;int a,b,c;
	Rprintf("<<%d>>\n",pos);c=0;
	for(a=0;a<dag->n;a++) {
	  for(b=0;b<dag->n;b++) 
	    if(bit_test(q,b)) { Rprintf("%d ",b+1);c++; }
	  r += bitstr_size(dag->n);
	  if(c) Rprintf("-> %d\n",a+1);
	}
	error("Invalid DAG sampled?!");
      }
      pos++;
    }
    cur->c_close(cur);
    SET_SLOT(ret,install("vertices"),ScalarInteger(dag->n));
    SET_SLOT(ret,install("dags"),reps);
    SET_SLOT(ret,install("p"),scores);
    SET_SLOT(ret,install("n"),sampled);
    SET_SLOT(ret,install("o"),overlap);
    UNPROTECT(4);
  }
  PutRNGstate();
  db->close(db,0);
  vmaxset(vmax);
  UNPROTECT(1);
  return ret;
}



/** Importance Sampling of Dags **/

SEXP BNSL_ISampleDagFromOrder(SEXP d,SEXP max_parents,SEXP order,SEXP N) {
  SEXP ret;
  char *vmax = vmaxget();
  dag_ctx   dag  = bnsl_dag_ctx(d,INTEGER(max_parents)[0]);
  pset_ctx *pset = (pset_ctx*)R_alloc(sizeof(pset_ctx),dag->n);
  double   *w    = (double*)R_alloc(sizeof(double),INTEGER(N)[0]);
  int       i;
  double    dags=1;
  
  double    eW,vW,ESS;

  /* Prepare the parent set list for each entry */
  for(i=0;i<dag->n;i++) {
    bnsl_dag_update_score(dag,i+1);
    pset[i] = bnsl_pset_ctx(d,INTEGER(max_parents)[0]);
  }
  for(i=0;i<dag->n;i++) {
    bnsl_pset_prepare(pset[i],INTEGER(order),i,INTEGER(order)[i]);
    dags   *= (double)pset[i]->used; 
  }
  Rprintf("%.0f dags possible\n",dags);

  GetRNGstate();
  PROTECT(ret = allocVector(REALSXP,INTEGER(N)[0]));
  for(i=0;i<INTEGER(N)[0];i++) {
    /*    w[i]         =  ImportanceSampleDag(pset,dag);*/
    REAL(ret)[i] = -bnsl_dag_score(dag);
  }
  eW = logsumexp(w,INTEGER(N)[0])-log((double)INTEGER(N)[0]);
  for(i=0;i<INTEGER(N)[0];i++) {
    double x = w[i]-eW;
    w[i] = log(1.0 + exp(2.0*x)-exp(x+log(2))) + 2.0*eW;
  }
  vW = logsumexp(w,INTEGER(N)[0])-log(-1.0 + (double)INTEGER(N)[0]);
  ESS = 100.0/(1.0 + exp(vW-2.0*eW));
  Rprintf("ESS = %.2f%\n",ESS);
  vmaxset(vmax);
  PutRNGstate();
  UNPROTECT(1);
  return ret;
}

SEXP BNSL_SampleDagFromOrder(SEXP d,SEXP max_parents,SEXP order,SEXP N) {
  SEXP  ret;
  char *vmax = vmaxget();
  dag_ctx dag= bnsl_dag_ctx(d,INTEGER(max_parents)[0]);
  move *moves = (move*)R_alloc(sizeof(move),dag->n*dag->n);
  int   i,j,t1,t2,a,r;
  double score,score_new,mh;
  int    o_n = length(order);
  int*   o   = INTEGER(order);


  GetRNGstate();
  PROTECT(ret = allocVector(REALSXP,INTEGER(N)[0]));
  a = r = 0;
  bnsl_dag_clear(dag);
  for(i=0;i<o_n;i++) {
    int max_k = unif_int(i<dag->k?i+1:dag->k+1);
    if(max_k==0) {
      bnsl_dag_update_score(dag,o[i]);
    } else
      while(max_k>0) {
	j = unif_int(i);
	if(!bnsl_dag_has_edge(dag,o[j],o[i])) {
	  bnsl_dag_set_edge(dag,o[j],o[i],1);
	  max_k--;
	}
      }
  }
  score = -bnsl_dag_score(dag);
  t1 = count_moves(dag,o,moves);

  for(i=0;i<INTEGER(N)[0];i++) {
    if(StepDag(dag,o,moves,&score,&t1))
      a++; else r++;
    REAL(ret)[i] = score;
  }
  Rprintf("%d %d\n",a,r);
  vmaxset(vmax);
  UNPROTECT(1);
  PutRNGstate();
  return ret;
}

SEXP BNSL_FeatureScoreFromDagSample(SEXP dagSample) {
  SEXP res,dags,sampled,overlap;
  int  i,j,n       = INTEGER(GET_SLOT(dagSample,install("vertices")))[0];
  double total,inc;
  PROTECT(res = allocVector(REALSXP,n*n));
  dags    = GET_SLOT(dagSample,install("dags"));
  sampled = GET_SLOT(dagSample,install("n"));
  overlap = GET_SLOT(dagSample,install("o"));
  
  memset(REAL(res),0,sizeof(double)*n*n);
  for(i=0;i<length(dags);i++) {
    SEXP   dag = VECTOR_ELT(dags,i);
    double inc = ((double)REAL(sampled)[i])/((double)REAL(overlap)[i]);
    /*    double inc = ((double)REAL(sampled)[i]);*/
    for(j=0;j<n*n;j++)
      if(bit_test(RAW(dag),j)) REAL(res)[j] += inc;
    /*    total += inc;*/
  }
  /*  for(i=0;i<n*n;i++) REAL(res)[i] /= total;*/
  UNPROTECT(1);
  return res;
}

