#include <R_ext/RS.h>
#include <Rinternals.h>
#include <Rmath.h>
#include <sys/types.h>
#include <limits.h>
#include <db.h>
#include <fenv.h>
#include "bitstring.h"

extern double logsumexp(double *,int); /* util.c */
extern void get_db_info(SEXP,DB**,char**,int*);
extern int      twiddle(int* x,int* y,int* z,int* p);
extern void inittwiddle(int,int,int*);

#define GET_SCORE(v) { \
 int r,s; \
 if((r = db->get(db,NULL,key,data,0)) != 0) {  \
   Rprintf("%d :",extension); \
   for(s=0;s<n;s++) if(bit_test(pa,s)) Rprintf("%d ",s+1); \
   Rprintf("\n"); \
   error("key unknown"); \
 }\
 v = ((double*)data->data)[0]; \
}

double score_extension_max(DB*db,DBT*key,DBT*data,
			   int*order,int o_n,
			   int extension,int max_parents,int n) {
  int    *P,x,y,z,i,j,p,t;
  int    m_p = (o_n > max_parents) ? max_parents : o_n;
  double best_score;
  char*  pa = key->data+1;
  memset(key->data,0,key->size);
  ((char*)key->data)[0] = extension;
  GET_SCORE(best_score);
  if(o_n == 1) {
    double result;
    bit_set(pa,order[0]-1);
    GET_SCORE(result);
    if(result > best_score) best_score = result;
    bit_clear(pa,order[0]-1);
  }
  if(o_n < 2) return best_score;
  
  P = (int*)R_alloc(o_n+2,sizeof(double));
  for(p=m_p;p>=1;p--) {
    double result;
    memset(pa,0,key->size-1);
    for(i=o_n-p;i<o_n;i++) bit_set(pa,order[i]-1);
    inittwiddle(p,o_n,P);
    GET_SCORE(result);if(best_score < result) result = best_score;
    while(!twiddle(&x,&y,&z,P)) {
      bit_set(pa,order[x]-1);
      bit_clear(pa,order[y]-1);
      GET_SCORE(result);if(best_score < result) result = best_score;
    }
  }
  return best_score;
}

double score_extension(DB*db,DBT*key,DBT*data,
		       int*order,int o_n,
		       int extension,int max_parents,
		       double*scores,int score_max,int n) {
  int     m_p = (o_n > max_parents) ? max_parents : o_n;
  char   *pa = key->data+1;
  int    *P,x,y,z,i,j,p,t;
  double temp;
  double result = 0.0;
  memset(key->data,0,key->size);
  ((char*)key->data)[0] = extension;  



  feclearexcept (FE_ALL_EXCEPT);
  GET_SCORE(result);
  scores[0] = result;t=1;
  if(o_n == 1) {
    bit_set(pa,order[0]-1);
    GET_SCORE(temp);
    if(result > temp) 
      result = result + log1p(exp(temp-result));
    else
      result = temp + log1p(exp(result-temp));
    bit_clear(pa,order[0]-1);
  }
  if(o_n < 2) {
    if(fetestexcept(FE_UNDERFLOW) || isnan(result)) {
      return (scores[0] > temp) ? scores[0] : temp;
    }
    return result;
  }
  
  P = (int*)R_alloc(o_n+2,sizeof(double));
  for(p=m_p;p>=1;p--) {		/* The more parents the smaller the scores */
    memset(pa,0,key->size-1);
    for(i=o_n-p;i<o_n;i++) bit_set(pa,order[i]-1);
    inittwiddle(p,o_n,P);
    GET_SCORE(result);		/* The initial score */
    while(!twiddle(&x,&y,&z,P)) {
      double temp;
      bit_set(pa,order[x]-1);
      bit_clear(pa,order[y]-1);
      GET_SCORE(temp);
      scores[t++] = temp;
    }
  }
  result = logsumexp(scores,t);
  /* Test the floating point exception here */
  /*  if(fetestexcept(FE_UNDERFLOW)) 
      warning("floating point underflow in logsumexp");*/
  return result;
}

double score_delta(DB*db,DBT*key,DBT*data,
		   int*order,int o_n,int candidate,
		   int extension,int max_parents,
		   double*scores,int score_max,int n) {
  int     m_p = (o_n > max_parents-1) ? max_parents-1 : o_n;
  char   *pa = key->data+1;
  double result = 0.0;
  int    *P,x,y,z,i,j,p,t;
  memset(key->data,0,key->size);
  ((char*)key->data)[0] = extension;  

  bit_set(pa,candidate-1);

  GET_SCORE(result);
  scores[0] = result;t=1;
  if(o_n == 1) {
    double temp;
    bit_set(pa,order[0]-1);
    GET_SCORE(temp);
    if(result > temp) 
      result = result + log1p(exp(temp-result));
    else
      result = temp + log1p(exp(result-temp));
    bit_clear(pa,order[0]-1);
  }
  if(o_n < 2) { bit_clear(pa,candidate-1);return result; }
  
  P = (int*)R_alloc(o_n+2,sizeof(double));


  for(p=m_p;p>=1;p--) {		/* The more parents the smaller the scores */
    memset(pa,0,key->size-1);
    for(i=o_n-p;i<o_n;i++) bit_set(pa,order[i]-1);
    inittwiddle(p,o_n,P);
    GET_SCORE(result);		/* The initial score */
    while(!twiddle(&x,&y,&z,P)) {
      double temp;
      bit_set(pa,order[x]-1);
      bit_clear(pa,order[y]-1);
      GET_SCORE(temp);
      scores[t++] = temp;
    }
  }
  bit_clear(pa,candidate-1);
  return logsumexp(scores,t);
}

SEXP BNSL_SISOMethod(SEXP _db,SEXP vars,SEXP M,SEXP T) {
  DB*  db;
  DBT  key,data;
  SEXP    order,result,names;
  int     i,on,score_max,n,t;
  double  iW,score;
  double  *scores;  

  int    *cand,cn,sel,slow;
  double *pdf,norm,cdf,ru,Temp;

  /* Try to do fast fp calculations */
  slow = 1;
  Temp = 1.0/REAL(T)[0];
  
  /* Setup the database */
  get_db_info(_db,&db,(char**)&(key.data),&n);
  key.size   = bitstr_size(n)+1;
  memset((char*)key.data,0,key.size);
  score_max = 1;
  for(i=1;i<=INTEGER(M)[0];i++)
    score_max += choose(length(vars),i);
  scores = (double*)R_alloc(score_max,sizeof(double));
  key.flags  = 0;
  data.flags = 0;
  /* Done Database Setup */

  /* Build the weights */
  PROTECT(result = allocVector(VECSXP,3));
  PROTECT(names  = allocVector(STRSXP,3));
  SET_STRING_ELT(names,0,mkChar("ordering"));
  SET_STRING_ELT(names,1,mkChar("weight"));
  SET_STRING_ELT(names,2,mkChar("score"));
  setAttrib(result,R_NamesSymbol,names);
  PROTECT(order = allocVector(INTSXP,length(vars)));
  SET_VECTOR_ELT(result,0,order);

  score = 0;
  iW    = 0;
  cn    = length(vars);
  cand = (int*)R_alloc(cn,sizeof(int));    /* candidates                 */
  memset(cand,0,sizeof(int)*cn);
  pdf  = (double*)R_alloc(cn,sizeof(double)); /* unnormalized probabilities */
  memset(pdf,0,sizeof(double)*cn);
  for(i=0;i<cn;i++) {
    cand[i] = INTEGER(vars)[i];
    pdf[i]  = Temp*score_extension(db,&key,&data,INTEGER(order),0,
				   cand[i],INTEGER(M)[0],scores,score_max,n);
  }

  GetRNGstate();
  on = 0;
  do {
    int temp;double max;
    revsort(pdf,cand,cn);
    norm = logsumexp(pdf,cn);	/* Normalizing constant */
    cdf  = 0.0;ru = unif_rand();
    /*    if(exp(pdf[0]-norm) > 1.0-REAL(rew)[0]) {
      for(i=0;i<cn;i++) {
	cdf += (i==0) ? 1.0-REAL(rew)[0] : REAL(rew)[0]/(double)(cn-1);
	if(ru < cdf) break;
      }
      iW += (i==0) ? log(1.0-REAL(rew)[0]) : log(REAL(rew)[0]/(double)(cn-1));
      } else { */
    for(i=0;i<cn;i++) {
      cdf += exp(pdf[i]-norm);
      if(ru < cdf) break;
    }
    iW    += pdf[i]/Temp;      
    /*    }*/
    
    /* Select an item */
    score += pdf[i];
    
    temp = cand[i];
    cand[i] = cand[cn-1];
    pdf[i]  = pdf[cn-1];cn--;

    if(!slow) {
      for(i=0;i<cn;i++) {
	double f;
	feclearexcept (FE_ALL_EXCEPT);

	f = score_delta(db,&key,&data,INTEGER(order),on,temp,
				    cand[i],INTEGER(M)[0],scores,score_max,n);
	pdf[i] = (f > pdf[i]) ? 
	  (double)(log1p(exp(pdf[i]-f))+f) : (double)(log1p(exp(f-pdf[i]))+pdf[i]);
	if(fetestexcept(FE_UNDERFLOW)) {
	  warning("switching to slow floating point methods");
	  slow = 1;
	  break;
	}
      }
    }
    INTEGER(order)[on++] = temp;
    if(slow) {
      for(i=0;i<cn;i++)
	pdf[i] = Temp*score_extension(db,&key,&data,INTEGER(order),on,
				 cand[i],INTEGER(M)[0],scores,score_max,n);
    }
  } while(cn > 0);
  PutRNGstate();
  SET_VECTOR_ELT(result,1,ScalarReal(iW));
  SET_VECTOR_ELT(result,2,ScalarReal(score));  
  UNPROTECT(3);
  return result;
}


SEXP BNSL_CalculateExtension(SEXP _db,SEXP order,SEXP ext,SEXP M,SEXP usemax) {
  DB*  db;
  DBT  key,data;
  int  i,n,o_n;
  int     score_max;
  double *scores;
  int    *combo;
  SEXP res;

  get_db_info(_db,&db,(char**)&(key.data),&n);
  key.size   = bitstr_size(n)+1;
  memset((char*)key.data,0,key.size);
  score_max = 1;
  for(i=1;i<=INTEGER(M)[0];i++)
    score_max += choose(length(order),i);
  scores = (double*)R_alloc(score_max,sizeof(double));

  key.flags  = 0;
  data.flags = 0;
  o_n = length(order);
  n   = length(ext);
  PROTECT(res = allocVector(REALSXP,length(ext)));
  for(i=0;i<n;i++)
    REAL(res)[i] = (LOGICAL(usemax)[0]) ?
      score_extension_max(db,&key,&data,INTEGER(order),o_n,
			  INTEGER(ext)[i],INTEGER(M)[0],n)
      : score_extension(db,&key,&data,
			INTEGER(order),o_n,
			INTEGER(ext)[i],INTEGER(M)[0],
			scores,score_max,n);
  UNPROTECT(1);
  return res;
}

SEXP BNSL_ScoreOrdering(SEXP _db,SEXP order,SEXP M,SEXP usemax) {
  DB*  db;
  DBT  key,data;
  int  i,n,o_n;
  int     score_max;
  double *scores,value;
  int    *combo;
  SEXP res;

  /* Setup the database */
  get_db_info(_db,&db,(char**)&(key.data),&n);
  key.size   = bitstr_size(n)+1;
  memset((char*)key.data,0,key.size);
  score_max = 1;
  for(i=1;i<=INTEGER(M)[0];i++)
    score_max += choose(length(order),i);
  scores = (double*)R_alloc(score_max,sizeof(double));
  key.flags  = 0;
  data.flags = 0;
  
  value = 0.0;
  for(i=0;i<length(order);i++)
    value += (LOGICAL(usemax)) ?
      score_extension_max(db,&key,&data,INTEGER(order),i,INTEGER(order)[i],
			  INTEGER(M)[0],n)
      : score_extension(db,&key,&data,
			INTEGER(order),i,INTEGER(order)[i],
			INTEGER(M)[0],scores,score_max,n);
  return ScalarReal(value);
}

