#include <bitstring.h>
#include "common.h"
#include <Rdefines.h>
#include <db.h>
#include "scoring.h"

#define MOVE_ADD 0
#define MOVE_REM 1
#define MOVE_REV 2
#define MAKE_MOVE(I,J,TYPE) ((I) << 16)|((J) << 8)|(TYPE))
#define DAG_I(X) (((X) >> 16)&0xFF)
#define DAG_J(X) (((X) >>  8)&0xFF)
#define DAG_T(X) ((X)&0xFF)
#define Min(A,B) (((A)<(B))?(A):(B))

extern SEXP OrderRandom(int);

typedef struct _gmove {
  int  i,j;
  char move;
} gmove;

int dag_count_moves(dag_ctx dag,gmove*moves) {
  int i,j;
  int count;
  count = 0;
  for(i=0;i<dag->n;i++) {
    for(j=0;j<dag->n;j++) {
      if(i != j) {

	if(bnsl_dag_add_edge(dag,i+1,j+1)) {
	  moves[count].i = i+1;moves[count].j = j+1;moves[count].move=1;
	  count++;
	  bnsl_dag_remove_edge(dag,i+1,j+1);
	}
	if(bnsl_dag_remove_edge(dag,i+1,j+1)) {
	  moves[count].i = i+1;moves[count].j = j+1;moves[count].move=2;
	  count++;
	  bnsl_dag_add_edge(dag,i+1,j+1);
	}
	if(bnsl_dag_reverse_edge(dag,i+1,j+1)) {
	  moves[count].i = i+1;moves[count].j = j+1;moves[count].move=3;
	  count++;
	  bnsl_dag_reverse_edge(dag,j+1,i+1);
	}

      }
    }
  }
  return count;
}

double BayesFactor(dag_ctx old,dag_ctx dag,int i,int j,int rev) {
  double bf1 = exp(bnsl_dag_get_score(dag,j)-bnsl_dag_get_score(old,j));
  double bf2 = exp(bnsl_dag_get_score(dag,i)-bnsl_dag_get_score(old,i));
  /*
  Rprintf("%d %.2f %.2f %.2f,%.2f\n",j,
	  bnsl_dag_get_score(dag,j),
	  bnsl_dag_get_score(old,j),bf1,bf2);
  */
  return (rev)?bf1*bf2:bf1;
}

void DAGRandomChain(dag_ctx dag) {
  SEXP o;
  bnsl_dag_clear(dag);
  PROTECT(o = OrderRandom(dag->n));
  
  UNPROTECT(1);
}


SEXP BNSL_StructureEE(SEXP d,SEXP max_parents,
		      SEXP _N,SEXP xH,SEXP xT,SEXP X) {
  
}



SEXP BNSL_StructureMCMC(SEXP d,SEXP max_parents,
			SEXP N,SEXP thin,SEXP start) {
  char     *vmax = vmaxget();
  dag_ctx   dag  = bnsl_dag_ctx(d,INTEGER(max_parents)[0]);
  dag_ctx   old  = bnsl_dag_ctx(d,INTEGER(max_parents)[0]);
  SEXP      res,ret,dags;
  int       i,q,add,sub,rev,acc,rej;
  double    score,moves;
  gmove    *movelist,*movelist2;


  GetRNGstate();
  PROTECT(res = allocVector(REALSXP,INTEGER(N)[1]));
  PROTECT(dags= allocVector(VECSXP,INTEGER(N)[1]));
  movelist = (gmove*)R_alloc(sizeof(struct _gmove),2*dag->n*dag->n);
  movelist2= (gmove*)R_alloc(sizeof(struct _gmove),2*dag->n*dag->n);
  /* Initialize a random chain dag */
  if(start == R_NilValue) {
    int  i;
    SEXP o;
    PROTECT(o = OrderRandom(dag->n));
    for(i=0;i<dag->n-1;i++) {
      bnsl_dag_add_edge(dag,INTEGER(o)[i],INTEGER(o)[i+1]);
      bnsl_dag_update_score(dag,INTEGER(o)[i+1]);
    }
    bnsl_dag_update_score(dag,INTEGER(o)[0]);
    UNPROTECT(1);
  } else {
    unsigned char *r = RAW(start);
    int   i,j;
    for(i=0;i<dag->n;i++) {
      for(j=0;j<dag->n;j++) 
	bnsl_dag_set_edge_no(dag,j+1,i+1,bit_test(r,j)?1:0);
      bnsl_dag_update_score(dag,i+1);
      r += bitstr_size(dag->n);
    }
  }

  score = bnsl_dag_score(dag);
  moves = dag_count_moves(dag,movelist);
  add = sub = rev = acc = rej = 0;
  q = 0;
  for(i=0;i<INTEGER(N)[0]+INTEGER(N)[1];q++) {
    double new_score,u,mh,new_moves;
    int    move  = unif_int((int)moves);
    bnsl_dag_copy(old,dag);
    switch(movelist[move].move) {
    case 1:
      bnsl_dag_add_edge(dag,movelist[move].i,movelist[move].j);
      bnsl_dag_update_score(dag,movelist[move].j);
      mh = BayesFactor(old,dag,movelist[move].i,movelist[move].j,0);
      add++;
      break;
    case 2:
      bnsl_dag_remove_edge(dag,movelist[move].i,movelist[move].j);
      bnsl_dag_update_score(dag,movelist[move].j);
      mh = BayesFactor(old,dag,movelist[move].i,movelist[move].j,0);
      sub++;
      break;
    case 3:
      bnsl_dag_reverse_edge(dag,movelist[move].i,movelist[move].j);
      bnsl_dag_update_score(dag,movelist[move].j);
      bnsl_dag_update_score(dag,movelist[move].i);
      mh = BayesFactor(old,dag,movelist[move].i,movelist[move].j,1);
      rev++;
      break;
    }

    new_score = bnsl_dag_score(dag);
    new_moves = dag_count_moves(dag,movelist2);
    /* Possibly Reject the move */
    u = unif_rand();
    /*    Rprintf("%.2f < %.2f (%.2f %.2f) %c\n",u,mh,score,new_score,
	  u<Min(mh*(moves/new_moves),1)?'+':'-');*/
    mh = Min(mh*(moves/new_moves),1);
    if(u < mh) {
      gmove* temp;
      score     = new_score;
      moves     = new_moves;
      temp      = movelist;
      movelist  = movelist2;
      movelist2 = temp;
      acc++;
    } else {
      bnsl_dag_copy(dag,old);
      rej++;
    }
    if(i>=INTEGER(N)[0]) {
      REAL(res)[i-INTEGER(N)[0]] = -score;
      SEXP d;
      PROTECT(d = allocVector(RAWSXP,dag->n*bitstr_size(dag->n)));
      memcpy(RAW(d),dag->dag,sizeof(char)*dag->n*bitstr_size(dag->n));
      SET_VECTOR_ELT(dags,i,d);
      UNPROTECT(1);
    }
    if(q % INTEGER(thin)[0] == 0) i++;
  }
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("sample")));
  SET_SLOT(ret,install("score"),res);
  SET_SLOT(ret,install("population"),dags);
  SET_SLOT(ret,install("temperature"),ScalarReal(1.0));
  SET_SLOT(ret,install("energy"),ScalarReal(R_NegInf));
  PutRNGstate();
  bnsl_dag_ctx_done(dag);
  vmaxset(vmax);
  UNPROTECT(3);
  Rprintf("%d add moves, %d subtract moves, %d reverse moves (%.2f)\n",
	  add,sub,rev,(double)100.0*acc/(double)(acc+rej));
  return ret;
}


SEXP BNSL_DagMakeSample(SEXP dags) {
  DB *db;DBT key,data;DBC*cur;
  char *vmax = vmaxget();
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_HASH,DB_CREATE,0);
  memset(&key,0,sizeof(key));
  memset(&data,0,sizeof(data));
  int i,count,unique=0;
  for(i=0;i<length(dags);i++) {
    SEXP d= VECTOR_ELT(dags,i);
    key.data = RAW(d);
    key.size = length(d);
    if(db->get(db,NULL,&key,&data,0)==0) {
      ((int*)data.data)[0]++;
    } else {
      data.data = &count;count = 1;
      data.size = sizeof(int);
      unique++;
    }
    db->put(db,NULL,&key,&data,0);
  }
  SEXP ret,d,n,o;
  PROTECT(ret = NEW_OBJECT(MAKE_CLASS("dagSample")));
  PROTECT(d   = allocVector(VECSXP,unique));
  PROTECT(n   = allocVector(INTSXP,unique));
  PROTECT(o   = allocVector(INTSXP,unique));
  db->cursor(db,NULL,&cur,0);
  i=0;
  while(cur->c_get(cur,&key,&data,DB_NEXT)==0) {
    SEXP o;
    PROTECT(o = allocVector(RAWSXP,key.size));
    memcpy(RAW(o),key.data,key.size);
    INTEGER(n)[i] = ((int*)data.data)[0];
    INTEGER(o)[i] = 1;
    i++;
  }
  cur->c_close(cur);
  SET_SLOT(ret,install("dags"),d);
  SET_SLOT(ret,install("n"),n);
  SET_SLOT(ret,install("o"),o);
  db->close(db,0);
  vmaxset(vmax);
  UNPROTECT(4);
  return ret;
}
