#include "common.h"
#include "fib.h"
#include <db.h>

typedef struct oinfo {
  int   n;			/* Steps */
  float g;			/* A-Star heuristic function */
  char order[0];		/* Ordering */
} *oinfo;

/* Used as a heuristic score for our A-star search */
float inline hamming_dist(char *o1,char *o2,int n) {
  int   i;
  float f = 0.0;
  for(i=0;i<n;i++) if(o1[i]!=o2[i]) f += 1.0;
  return f;
}

/* An A-star comparison */
int oinfo_cmp(void*_x,void*_y) {
  float fx = (float)((oinfo)_x)->n + ((oinfo)_x)->g;
  float fy = (float)((oinfo)_y)->n + ((oinfo)_y)->g;
  if(fx< fy) return -1;
  if(fx==fy) return  0;
  return 1;
}

oinfo inline new_oinfo(int l) {
  return (oinfo)R_alloc(sizeof(struct oinfo)+sizeof(char)*l,1);
}
oinfo inline dup_oinfo(int l,oinfo src) {
  oinfo a = new_oinfo(l);
  memcpy(a,src,sizeof(struct oinfo)+sizeof(char)*l);
  a->n = 0;
  a->g = 0.0;
  return a;
}

void inline insert_moves(struct fibheap *a,char *dest,int l,
			 oinfo node,int (*can_insert)(char*)) {
  int i,j,n;
  char  t;
  n = node->n+1;
  /*  Rprintf("%d %.0f\n",node->n,node->g); */
  for(i=1;i<l;i++) {
    for(j=0;j<i;j++) {
      /* Make swap */
      t = node->order[i];
      node->order[i] = node->order[j];
      node->order[j] = t;
      /* Maybe insert swap */
      if(can_insert(node->order)) {
	node->n = n;
	node->g = hamming_dist(node->order,dest,l);
	fh_insert(a,(void*)node);
	node = dup_oinfo(l,node);
      }
      /* Undo swap */
      t = node->order[i];
      node->order[i] = node->order[j];
      node->order[j] = t;
    }
  }
}

SEXP BNSL_SwapSearch(SEXP start,SEXP finish) {
  oinfo           min;
  DB*             db;
  DBT             key,data;
  int             i,l,n,c;
  char           *dest;
  char           *vmax = vmaxget();
  char            t,d=1;
  struct fibheap *a;
  SEXP            res;
  
  l = length(start);
  if(l != length(finish))
    error("Source and Destination orders are different lengths");
  dest = (char*)R_alloc(sizeof(char),l);
  for(i=0;i<l;i++) dest[i] = INTEGER(finish)[i];

  /* Create a TABU hash */
  db_create(&db,NULL,0);
  db->open(db,NULL,NULL,NULL,DB_HASH,DB_CREATE,0);
  memset(&key, 0,sizeof(DBT));
  memset(&data,0,sizeof(DBT));
  key.size  = sizeof(char)*l;
  data.size = sizeof(char);
  data.data = &d;
  int can_insert(char*a) {
    key.data = a;
    if(db->get(db,NULL,&key,&data,0)!=0) {
      data.data = &d;data.size = sizeof(char);
      db->put(db,NULL,&key,&data,0);
      n++;
      return 1;
    } else c++;
    return 0;
  }
  PROTECT(res = allocVector(REALSXP,2));
  n = c = 0;
  min = new_oinfo(l);
  for(i=0;i<l;i++) min->order[i] = INTEGER(start)[i];
  min->n = 0;
  REAL(res)[1] = min->g = hamming_dist(min->order,dest,l);
  a = fh_makeheap();
  fh_setcmp(a,oinfo_cmp);
  do {
    if(min->g == 0) break;
    insert_moves(a,dest,l,min,can_insert);
    min = (oinfo)fh_extractmin(a);
  } while(min);

  if(min == NULL)
    error("Unable to locate a solution");
  Rprintf("%d steps and %d clashes (%d)\n",n,c,n+c);
  REAL(res)[0] = min->n;
  fh_deleteheap(a);
  db->close(db,0);
  vmaxset(vmax);
  UNPROTECT(1);
  return res;
}
